"""The use command for the Mirascope CLI."""
import os

from typer import Argument

from ...enums import MirascopeCommand
from ..constants import CURRENT_REVISION_KEY
from ..utils import (
    check_status,
    find_prompt_path,
    get_user_mirascope_settings,
    parse_prompt_file_name,
    prompts_directory_files,
    run_format_command,
    update_version_text_file,
    write_prompt_to_template,
)


def use_command(
    prompt_file_name: str = Argument(
        help="Prompt file to use",
        autocompletion=prompts_directory_files,
        parser=parse_prompt_file_name,
    ),
    version: str = Argument(
        help="Version of prompt to use",
    ),
) -> None:
    """Uses the version and prompt specified by the user.

    The contents of the prompt in the versions directory are copied to the user's
    prompts directory, based on the version specified by the user. The version file is
    updated with the new revision.

    Args:
        prompt_file_name: The name of the prompt file to use.
        version: The version of the prompt file to use.

    Raises:
        FileNotFoundError: If the file is not found in the versions directory.
    """
    mirascope_settings = get_user_mirascope_settings()
    used_prompt_path = check_status(mirascope_settings, prompt_file_name)
    # Check status before continuing
    if used_prompt_path:
        print("Changes detected, please add or remove changes first.")
        print(f"\tmirascope add {prompt_file_name}".expandtabs(4))
        return
    version_directory_path = mirascope_settings.versions_location
    prompt_directory_path = mirascope_settings.prompts_location
    version_file_name = mirascope_settings.version_file_name
    prompt_versions_directory = os.path.join(version_directory_path, prompt_file_name)
    revision_file_path = find_prompt_path(prompt_versions_directory, version)
    version_file_path = os.path.join(prompt_versions_directory, version_file_name)
    if revision_file_path is None:
        raise FileNotFoundError(
            f"Prompt version {version} not found in {prompt_versions_directory}"
        )
    # Open versioned prompt file
    with open(revision_file_path, "r", encoding="utf-8") as file:
        content = file.read()
    # Write to user's prompt file
    prompt_file_path = os.path.join(prompt_directory_path, f"{prompt_file_name}.py")
    with open(prompt_file_path, "w+", encoding="utf-8") as file2:
        file2.write(write_prompt_to_template(content, MirascopeCommand.USE))
    if prompt_file_path:
        run_format_command(prompt_file_path)

    # Update version file with new current revision
    keys_to_update = {
        CURRENT_REVISION_KEY: version,
    }
    update_version_text_file(version_file_path, keys_to_update)

    print(f"Using {revision_file_path}")



"""A module for interacting with Mistral models."""
from .calls import MistralCall
from .extractors import MistralExtractor
from .tools import MistralTool
from .types import MistralCallParams, MistralCallResponse, MistralCallResponseChunk



"""A module for calling OpenAI's Chat Completion models."""
import datetime
from typing import Any, AsyncGenerator, ClassVar, Generator

from openai import AsyncOpenAI, OpenAI
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)

from ..base import BaseCall
from ..enums import MessageRole
from .tools import OpenAITool
from .types import OpenAICallParams, OpenAICallResponse, OpenAICallResponseChunk


class OpenAICall(BaseCall[OpenAICallResponse, OpenAICallResponseChunk, OpenAITool]):
    """A base class for calling OpenAI's Chat Completion models.

    Example:

    ```python
    from mirascope.openai import OpenAICall


    class BookRecommender(OpenAICall):
        prompt_template = "Please recommend a {genre} book"

        genre: str

    response = BookRecommender(genre="fantasy").call()
    print(response.content)
    #> There are many great books to read, it ultimately depends...
    ```
    """

    call_params: ClassVar[OpenAICallParams] = OpenAICallParams()

    def messages(self) -> list[ChatCompletionMessageParam]:
        """Returns the template as a formatted list of messages."""
        message_type_by_role = {
            MessageRole.SYSTEM: ChatCompletionSystemMessageParam,
            MessageRole.USER: ChatCompletionUserMessageParam,
            MessageRole.ASSISTANT: ChatCompletionAssistantMessageParam,
            MessageRole.TOOL: ChatCompletionToolMessageParam,
        }
        return [
            message_type_by_role[MessageRole(message["role"])](
                role=message["role"], content=message["content"]
            )
            for message in self._parse_messages(list(message_type_by_role.keys()))
        ]

    def call(self, **kwargs: Any) -> OpenAICallResponse:
        """Makes a call to the model using this `OpenAICall` instance.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            A `OpenAICallResponse` instance.

        Raises:
            OpenAIError: raises any OpenAI errors, see:
                https://platform.openai.com/docs/guides/error-codes/api-errors
        """
        kwargs, tool_types = self._setup(kwargs, OpenAITool)
        client = OpenAI(api_key=self.api_key, base_url=self.base_url)
        if self.call_params.wrapper is not None:
            client = self.call_params.wrapper(client)
        start_time = datetime.datetime.now().timestamp() * 1000
        completion = client.chat.completions.create(
            messages=self.messages(),
            stream=False,
            **kwargs,
        )
        return OpenAICallResponse(
            response=completion,
            tool_types=tool_types,
            start_time=start_time,
            end_time=datetime.datetime.now().timestamp() * 1000,
        )

    async def call_async(self, **kwargs: Any) -> OpenAICallResponse:
        """Makes an asynchronous call to the model using this `OpenAICall`.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            An `OpenAICallResponse` instance.

        Raises:
            OpenAIError: raises any OpenAI errors, see:
                https://platform.openai.com/docs/guides/error-codes/api-errors
        """
        kwargs, tool_types = self._setup(kwargs, OpenAITool)
        client = AsyncOpenAI(api_key=self.api_key, base_url=self.base_url)
        if self.call_params.wrapper_async is not None:
            client = self.call_params.wrapper_async(client)
        start_time = datetime.datetime.now().timestamp() * 1000
        completion = await client.chat.completions.create(
            messages=self.messages(),
            stream=False,
            **kwargs,
        )
        return OpenAICallResponse(
            response=completion,
            tool_types=tool_types,
            start_time=start_time,
            end_time=datetime.datetime.now().timestamp() * 1000,
        )

    def stream(self, **kwargs: Any) -> Generator[OpenAICallResponseChunk, None, None]:
        """Streams the response for a call using this `OpenAICall`.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Yields:
            A `OpenAICallResponseChunk` for each chunk of the response.

        Raises:
            OpenAIError: raises any OpenAI errors, see:
                https://platform.openai.com/docs/guides/error-codes/api-errors
        """
        kwargs, tool_types = self._setup(kwargs, OpenAITool)
        client = OpenAI(api_key=self.api_key, base_url=self.base_url)
        if self.call_params.wrapper is not None:
            client = self.call_params.wrapper(client)
        stream = client.chat.completions.create(
            messages=self.messages(),
            stream=True,
            **kwargs,
        )
        for chunk in stream:
            yield OpenAICallResponseChunk(chunk=chunk, tool_types=tool_types)

    async def stream_async(
        self, **kwargs: Any
    ) -> AsyncGenerator[OpenAICallResponseChunk, None]:
        """Streams the response for an asynchronous call using this `OpenAICall`.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Yields:
            A `OpenAICallResponseChunk` for each chunk of the response.

        Raises:
            OpenAIError: raises any OpenAI errors, see:
                https://platform.openai.com/docs/guides/error-codes/api-errors
        """
        kwargs, tool_types = self._setup(kwargs, OpenAITool)
        client = AsyncOpenAI(api_key=self.api_key, base_url=self.base_url)
        if self.call_params.wrapper_async is not None:
            client = self.call_params.wrapper_async(client)
        stream = await client.chat.completions.create(
            messages=self.messages(),
            stream=True,
            **kwargs,
        )
        async for chunk in stream:
            yield OpenAICallResponseChunk(chunk=chunk, tool_types=tool_types)



"""A module for interacting with Google's Gemini models."""
from .calls import GeminiCall
from .extractors import GeminiExtractor
from .tools import GeminiTool
from .types import GeminiCallParams, GeminiCallResponse, GeminiCallResponseChunk



"""Contains the schema for files created by the mirascope cli."""
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class MirascopeSettings(BaseModel):
    """Model for the user's mirascope settings."""

    mirascope_location: str
    versions_location: str
    prompts_location: str
    version_file_name: str
    format_command: Optional[str] = None
    auto_tag: Optional[bool] = None

    model_config = ConfigDict(extra="forbid")


class VersionTextFile(BaseModel):
    """Model for the version text file."""

    current_revision: Optional[str] = Field(default=None)
    latest_revision: Optional[str] = Field(default=None)


class MirascopeCliVariables(BaseModel):
    """Prompt version variables used internally by mirascope."""

    prev_revision_id: Optional[str] = Field(default=None)
    revision_id: Optional[str] = Field(default=None)


class ClassInfo(BaseModel):
    name: str
    bases: list[str]
    body: str
    decorators: list[str]
    docstring: Optional[str]


class FunctionInfo(BaseModel):
    name: str
    args: list[str]
    returns: Optional[str]
    body: str
    decorators: list[str]
    docstring: Optional[str]
    is_async: bool



"""The remove command for the Mirascope CLI."""
import os

from typer import Argument

from ..utils import (
    find_prompt_path,
    find_prompt_paths,
    get_prompt_versions,
    get_user_mirascope_settings,
    parse_prompt_file_name,
    prompts_directory_files,
)


def remove_command(
    prompt_file_name: str = Argument(
        help="Prompt file to remove",
        autocompletion=prompts_directory_files,
        parser=parse_prompt_file_name,
        default="",
    ),
    version: str = Argument(
        help="Version of prompt to use",
    ),
):
    """Removes the version from the versions directory

    All versions with prev_revision_id matching the deleted version are detached.

    Args:
        prompt_file_name: The name of the prompt file to remove.
        version: The version of the prompt to remove

    Raises:
        FileNotFoundError: If the file is not found in the specified prompts or
          versions directory.
    """
    mirascope_settings = get_user_mirascope_settings()
    version_directory_path = mirascope_settings.versions_location
    version_file_name = mirascope_settings.version_file_name
    prompt_versions_directory = os.path.join(version_directory_path, prompt_file_name)
    version_file_path = os.path.join(prompt_versions_directory, version_file_name)

    revisions = get_prompt_versions(version_file_path)
    if revisions.current_revision == version:
        print(
            f"Prompt {prompt_file_name} {version} is currently being used. "
            "Please switch to another version first."
        )
        return

    revision_file_path = find_prompt_path(prompt_versions_directory, version)
    if not revision_file_path:
        raise FileNotFoundError(
            f"Prompt version {version} not found in {prompt_versions_directory}"
        )
    # TODO: Implement rollback in case of failure
    os.remove(revision_file_path)

    # Detach any revisions that had the deleted version as their prev_revision_id
    revision_file_paths = find_prompt_paths(prompt_versions_directory, "")
    if revision_file_paths is None:
        revision_file_paths = []
    for revision_file_path in revision_file_paths:
        prev_revision_id_found = False
        with open(
            revision_file_path,
            "r",
            encoding="utf-8",
        ) as file:
            lines = file.readlines()
            for i, line in enumerate(lines):
                if "prev_revision_id" in line:
                    current_prev_revision_id = line.strip().split("=")[1].strip()
                    current_prev_revision_id = current_prev_revision_id[1:-1]
                    if current_prev_revision_id == version:
                        lines[i] = "prev_revision_id = None\n"
                        prev_revision_id_found = True
                    break
        if prev_revision_id_found:
            with open(revision_file_path, "w", encoding="utf-8") as file:
                file.writelines(lines)
            print(f"Detached {revision_file_path}")
    print(f"Prompt {prompt_file_name} {version} successfully removed")



"""mirascope package."""
import importlib.metadata
from contextlib import suppress

from .base import BaseCallParams, BasePrompt, Message, tags

with suppress(ImportError):
    from . import anthropic

with suppress(ImportError):
    from . import gemini

with suppress(ImportError):
    from . import openai

__version__ = importlib.metadata.version("mirascope")



"""A class for extracting structured information using Mistral chat models."""
import logging
from typing import Any, Generic, TypeVar

from ..base import BaseExtractor, ExtractedType
from .calls import MistralCall
from .tools import MistralTool

logger = logging.getLogger("mirascope")

T = TypeVar("T", bound=ExtractedType)


class MistralExtractor(BaseExtractor[MistralCall, MistralTool, T], Generic[T]):
    '''A class for extracting structured information using Mistral Chat models.

    Example:

    ```python
    from mirascope.mistral import MistralExtractor
    from pydantic import BaseModel
    from typing import Literal, Type

    class TaskDetails(BaseModel):
        title: str
        priority: Literal["low", "normal", "high"]
        due_date: str

    class TaskExtractor(MistralExtractor[TaskDetails]):
        extract_schema: Type[TaskDetails] = TaskDetails
        call_params = MistralCallParams(model="mistral-large-latest")

        prompt_template = """
        Prepare the budget report by next Monday. It's a high priority task.
        """


    task = TaskExtractor().extract(retries=3)
    assert isinstance(task, TaskDetails)
    print(task)
    # > title='Prepare the budget report' priority='high' due_date='next Monday'
    ```
    '''

    def extract(self, retries: int = 0, **kwargs: Any) -> T:
        """Extracts `extract_schema` from the Mistral call response.

        The `extract_schema` is converted into an `MistralTool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of Mistrals's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
            MistralException: raises any Mistral exceptions, see:
                https://github.com/mistralai/client-python/blob/main/src/mistralai/exceptions.py
        """
        return self._extract(MistralCall, MistralTool, retries, **kwargs)

    async def extract_async(self, retries: int = 0, **kwargs: Any) -> T:
        """Asynchronously extracts `extract_schema` from the Mistral call response.

        The `extract_schema` is converted into an `MistralTool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of Mistrals's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
            MistralException: raises any Mistral exceptions, see:
                https://github.com/mistralai/client-python/blob/main/src/mistralai/exceptions.py
        """
        return await self._extract_async(MistralCall, MistralTool, retries, **kwargs)



"""The add command for the Mirascope CLI."""
import os

from typer import Argument

from ...enums import MirascopeCommand
from ..constants import CURRENT_REVISION_KEY, LATEST_REVISION_KEY
from ..schemas import MirascopeCliVariables
from ..utils import (
    check_status,
    get_prompt_versions,
    get_user_mirascope_settings,
    parse_prompt_file_name,
    prompts_directory_files,
    run_format_command,
    update_version_text_file,
    write_prompt_to_template,
)


def add_command(
    prompt_file_name: str = Argument(
        help="Prompt file to add",
        autocompletion=prompts_directory_files,
        parser=parse_prompt_file_name,
        default="",
    ),
):
    """Adds the given prompt to the specified version directory.

    The contents of the prompt in the user's prompts directory are copied to the version
    directory with the next revision number, and the version file is updated with the
    new revision.

    Args:
        prompt_file_name: The name of the prompt file to add.

    Raises:
        FileNotFoundError: If the file is not found in the specified prompts directory.
    """
    mirascope_settings = get_user_mirascope_settings()
    version_directory_path = mirascope_settings.versions_location
    prompt_directory_path = mirascope_settings.prompts_location
    version_file_name = mirascope_settings.version_file_name
    auto_tag = mirascope_settings.auto_tag
    # Check status before continuing
    used_prompt_path = check_status(mirascope_settings, prompt_file_name)
    if not used_prompt_path:
        print("No changes detected.")
        return
    prompt_versions_directory = os.path.join(version_directory_path, prompt_file_name)

    # Check if prompt file exists
    if not os.path.exists(f"{prompt_directory_path}/{prompt_file_name}.py"):
        raise FileNotFoundError(
            f"Prompt {prompt_file_name}.py not found in {prompt_directory_path}"
        )
    # Create prompt versions directory if it doesn't exist
    if not os.path.exists(prompt_versions_directory):
        os.makedirs(prompt_versions_directory)
    version_file_path = os.path.join(prompt_versions_directory, version_file_name)
    versions = get_prompt_versions(version_file_path)

    # Open user's prompt file
    user_prompt_file = os.path.join(prompt_directory_path, f"{prompt_file_name}.py")
    with open(user_prompt_file, "r+", encoding="utf-8") as file:
        # Increment revision id
        if versions.latest_revision is None:
            # first revision
            revision_id = "0001"
        else:
            # default branch with incrementation
            latest_revision_id = versions.latest_revision
            revision_id = f"{int(latest_revision_id)+1:04}"
        # Create revision file
        revision_file = os.path.join(
            prompt_versions_directory, f"{revision_id}_{prompt_file_name}.py"
        )
        custom_variables = MirascopeCliVariables(
            prev_revision_id=versions.current_revision,
            revision_id=revision_id,
        )
        prompt_file = file.read()

        if auto_tag:
            new_prompt_file = write_prompt_to_template(
                prompt_file, MirascopeCommand.USE, custom_variables
            )
            # Replace contents of user's prompt file with new prompt file with tags
            file.seek(0)
            file.write(new_prompt_file)
            file.truncate()
            # Reset file pointer to beginning of file for revision file read
            file.seek(0)
            run_format_command(user_prompt_file)
        with open(
            revision_file,
            "w+",
            encoding="utf-8",
        ) as file2:
            file2.write(
                write_prompt_to_template(
                    prompt_file, MirascopeCommand.ADD, custom_variables
                )
            )
            keys_to_update = {
                CURRENT_REVISION_KEY: revision_id,
                LATEST_REVISION_KEY: revision_id,
            }
            update_version_text_file(version_file_path, keys_to_update)
    if revision_file:
        run_format_command(revision_file)
    print("Adding " f"{prompt_versions_directory}/{revision_id}_{prompt_file_name}.py")



"""A base abstract interface for extracting structured information using LLMs."""
import logging
from abc import ABC, abstractmethod
from enum import Enum
from inspect import isclass
from typing import (
    Annotated,
    Any,
    Callable,
    ClassVar,
    Generic,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    get_origin,
)

from pydantic import BaseModel, ConfigDict, ValidationError

from .calls import BaseCall
from .prompts import BasePrompt
from .tools import BaseTool, BaseType
from .types import BaseCallParams

logger = logging.getLogger("mirascope")

BaseModelT = TypeVar("BaseModelT", bound=BaseModel)
BaseCallT = TypeVar("BaseCallT", bound=BaseCall)
BaseToolT = TypeVar("BaseToolT", bound=BaseTool)
ExtractionType = Union[Type[BaseType], Type[BaseModel], Callable]
ExtractedType = Union[BaseType, BaseModelT, BaseToolT]
ExtractedTypeT = TypeVar("ExtractedTypeT", bound=ExtractedType)


def _is_base_type(type_: Any) -> bool:
    """Check if a type is a base type."""
    base_types = {str, int, float, bool, list, set, tuple}
    return (
        (isclass(type_) and issubclass(type_, Enum))
        or type_ in base_types
        or get_origin(type_) in base_types.union({Literal, Union, Annotated})
    )


class BaseExtractor(BasePrompt, Generic[BaseCallT, BaseToolT, ExtractedTypeT], ABC):
    """The base abstract interface for extracting structured information using LLMs."""

    extract_schema: ExtractionType

    api_key: ClassVar[Optional[str]] = None
    base_url: ClassVar[Optional[str]] = None
    call_params: ClassVar[BaseCallParams] = BaseCallParams[BaseToolT](
        model="gpt-3.5-turbo-0125"
    )

    @abstractmethod
    def extract(self, retries: int = 0) -> ExtractedTypeT:
        """Extracts the `extraction_schema` from an LLM call."""
        ...  # pragma: no cover

    @abstractmethod
    async def extract_async(self, retries: int = 0) -> ExtractedTypeT:
        """Asynchronously extracts the `extraction_schema` from an LLM call."""
        ...  # pragma: no cover

    ############################## PRIVATE METHODS ###################################

    def _extract(
        self,
        call_type: Type[BaseCallT],
        tool_type: Type[BaseToolT],
        retries: int,
        **kwargs: Any,
    ) -> ExtractedTypeT:
        """Extracts `extract_schema` from the call response.

        The `extract_schema` is converted into a tool, complete with a description of
        the tool, all of the fields, and their types. This allows us to take advantage
        of tools/function calling functionality to extract information from a prompt
        according to the context provided by the `BaseModel` schema.

        Args:
            tool_type: The type of tool to use for extraction. This enables shared code
                across various model providers that have slight variations but the same
                internal interfaces.
            retries: The number of call attempts to make on `ValidationError` before
                giving up and throwing the error to the user.
            **kwargs: Additional keyword arguments.

        Returns:
            An instance of the `extract_schema` with it's fields populated.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
        """
        kwargs, return_tool = self._setup(tool_type, kwargs)

        class TempCall(call_type):  # type: ignore
            prompt_template = self.prompt_template

            call_params = call_type.call_params

            model_config = ConfigDict(extra="allow")

        response = TempCall(**self.model_dump(exclude={"extract_schema"})).call(
            **kwargs
        )
        try:
            tool = response.tool
            if tool is None:
                raise AttributeError("No tool found in the completion.")
            if return_tool:
                return tool
            if _is_base_type(self.extract_schema):
                return tool.value
            model = self.extract_schema(**response.tool.model_dump())
            model._response = response  # type: ignore
            return model  # type: ignore
        except (AttributeError, ValueError, ValidationError) as e:
            if retries > 0:
                logging.info(f"Retrying due to exception: {e}")
                # TODO: include failure in retry prompt.
                return self._extract(call_type, tool_type, retries - 1, **kwargs)
            raise  # re-raise if we have no retries left

    async def _extract_async(
        self,
        call_type: Type[BaseCallT],
        tool_type: Type[BaseToolT],
        retries: int,
        **kwargs: Any,
    ) -> ExtractedTypeT:
        """Extracts `extract_schema` from the asynchronous call response.

        The `extract_schema` is converted into a tool, complete with a description of
        the tool, all of the fields, and their types. This allows us to take advantage
        of tools/function calling functionality to extract information from a prompt
        according to the context provided by the `BaseModel` schema.

        Args:
            tool_type: The type of tool to use for extraction. This enables shared code
                across various model providers that have slight variations but the same
                internal interfaces.
            retries: The number of call attempts to make on `ValidationError` before
                giving up and throwing the error to the user.
            **kwargs: Additional keyword arguments.

        Returns:
            An instance of the `extract_schema` with it's fields populated.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
        """
        kwargs, return_tool = self._setup(tool_type, kwargs)

        class TempCall(call_type):  # type: ignore
            prompt_template = self.prompt_template

            call_params = call_type.call_params

            model_config = ConfigDict(extra="allow")

        response = await TempCall(
            **self.model_dump(exclude={"extract_schema"})
        ).call_async(**kwargs)
        try:
            tool = response.tool
            if tool is None:
                raise AttributeError("No tool found in the completion.")
            if return_tool:
                return tool
            if _is_base_type(self.extract_schema):
                return tool.value
            model = self.extract_schema(**response.tool.model_dump())
            model._response = response  # type: ignore
            return model  # type: ignore
        except (AttributeError, ValueError, ValidationError) as e:
            if retries > 0:
                logging.info(f"Retrying due to exception: {e}")
                # TODO: include failure in retry prompt.
                return await self._extract_async(
                    call_type, tool_type, retries - 1, **kwargs
                )
            raise  # re-raise if we have no retries left

    def _setup(
        self, tool_type: Type[BaseToolT], kwargs: dict[str, Any]
    ) -> tuple[dict[str, Any], bool]:
        """Returns the call params kwargs and whether to return the tool directly."""
        call_params = self.call_params.model_copy(update=kwargs)
        kwargs = call_params.kwargs(tool_type=tool_type)
        if _is_base_type(self.extract_schema):
            tool = tool_type.from_base_type(self.extract_schema)  # type: ignore
            return_tool = False
        elif not isclass(self.extract_schema):
            tool = tool_type.from_fn(self.extract_schema)
            return_tool = True
        elif not issubclass(self.extract_schema, tool_type):
            tool = tool_type.from_model(self.extract_schema)
            return_tool = False
        else:
            tool = self.extract_schema
            return_tool = True
        kwargs["tools"] = [tool]
        return kwargs, return_tool



"""Type classes for interacting with Anthropics's Claude API."""
import xml.etree.ElementTree as ET
from typing import Any, Callable, Literal, Optional, Type, Union, cast

from anthropic import Anthropic, AsyncAnthropic
from anthropic._types import Body, Headers, Query
from anthropic.types import (
    ContentBlockDeltaEvent,
    ContentBlockStartEvent,
    Message,
    MessageStreamEvent,
)
from anthropic.types.completion_create_params import Metadata
from httpx import Timeout
from pydantic import ConfigDict

from ..base.types import BaseCallParams, BaseCallResponse, BaseCallResponseChunk
from .tools import AnthropicTool


class AnthropicCallParams(BaseCallParams[AnthropicTool]):
    """The parameters to use when calling d Claud API with a prompt.

    Example:

    ```python
    from mirascope.anthropic import AnthropicCall, AnthropicCallParams


    class BookRecommender(AnthropicCall):
        prompt_template = "Please recommend some books."

        call_params = AnthropicCallParams(
            model="anthropic-3-opus-20240229",
        )
    ```
    """

    max_tokens: int = 1000
    model: str = "claude-3-sonnet-20240229"
    metadata: Optional[Metadata] = None
    stop_sequences: Optional[list[str]] = None
    system: Optional[str] = None
    temperature: Optional[float] = None
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    extra_headers: Optional[Headers] = None
    extra_query: Optional[Query] = None
    extra_body: Optional[Body] = None
    timeout: Optional[Union[float, Timeout]] = 600

    wrapper: Optional[Callable[[Anthropic], Anthropic]] = None
    wrapper_async: Optional[Callable[[AsyncAnthropic], AsyncAnthropic]] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def kwargs(
        self,
        tool_type: Optional[Type[AnthropicTool]] = None,
        exclude: Optional[set[str]] = None,
    ) -> dict[str, Any]:
        """Returns the keyword argument call parameters."""
        extra_exclude = {"wrapper", "wrapper_async"}
        exclude = extra_exclude if exclude is None else exclude.union(extra_exclude)
        return super().kwargs(tool_type, exclude)


class AnthropicCallResponse(BaseCallResponse[Message, AnthropicTool]):
    """Convenience wrapper around the Anthropic Claude API.

    When using Mirascope's convenience wrappers to interact with Anthropic models via
    `AnthropicCall`, responses using `Anthropic.call()` will return an
    `AnthropicCallResponse`, whereby the implemented properties allow for simpler syntax
    and a convenient developer experience.

    Example:

    ```python
    from mirascope.anthropic import AnthropicCall


    class BookRecommender(AnthropicCall):
        prompt_template = "Please recommend some books."


    print(BookRecommender().call())
    ```
    """

    @property
    def tools(self) -> Optional[list[AnthropicTool]]:
        """Returns the tools for the 0th choice message."""
        if (
            not self.tool_types
            or self.response.stop_reason != "stop_sequence"
            or self.response.stop_sequence != "</function_calls>"
        ):
            return None

        try:
            root_node = ET.fromstring(
                f"<wrapper>{self.response.content}</function_calls></wrapper>"
            )
        except ET.ParseError as e:
            raise ValueError("Unable to parse tools from response") from e

        # There must be a <function_calls> tag since we successfully parsed the
        # XML with the manually added </function_calls> tag.
        tool_calls_node = cast(ET.Element, root_node.find("function_calls"))

        extracted_tools = []
        for tool_call_node in tool_calls_node:
            for tool_type in self.tool_types:
                tool_name_node = tool_call_node.find("tool_name")
                if (
                    tool_name_node is not None
                    and tool_name_node.text == tool_type.__name__
                    and (parameters := tool_call_node.find("parameters"))
                ):
                    tool = tool_type.from_tool_call(parameters)
                    extracted_tools.append(tool)
                    break

        return extracted_tools

    @property
    def tool(self) -> Optional[AnthropicTool]:
        """Returns the 0th tool for the 0th choice message."""
        tools = self.tools
        if tools:
            return tools[0]
        return None

    @property
    def content(self) -> str:
        """Returns the string content of the 0th message."""
        return self.response.content[0].text

    def dump(self) -> dict[str, Any]:
        """Dumps the response to a dictionary."""
        return {
            "start_time": self.start_time,
            "end_time": self.end_time,
            "output": self.response.model_dump(),
        }


class AnthropicCallResponseChunk(
    BaseCallResponseChunk[MessageStreamEvent, AnthropicTool]
):
    """Convenience wrapper around the Anthropic API streaming chunks.

    When using Mirascope's convenience wrappers to interact with Anthropic models via
    `AnthropicCall`, responses using `AnthropicCall.stream()` will yield
    `AnthropicCallResponseChunk`, whereby the implemented properties allow for simpler
    syntax and a convenient developer experience.

    Example:

    ```python
    from mirascope.anthropic import AnthropicCall


    class BookRecommender(AnthropicCall):
        prompt_template = "Please recommend some books."


    for chunk in BookRecommender().stream():
        print(chunk, end="")
    ```
    """

    @property
    def type(
        self,
    ) -> Literal[
        "message_start",
        "message_delta",
        "message_stop",
        "content_block_start",
        "content_block_delta",
        "content_block_stop",
    ]:
        """Returns the type of the chunk."""
        return self.chunk.type

    @property
    def content(self) -> str:
        """Returns the string content of the 0th message."""
        if isinstance(self.chunk, ContentBlockStartEvent):
            return self.chunk.content_block.text
        if isinstance(self.chunk, ContentBlockDeltaEvent):
            return self.chunk.delta.text
        return ""



"""A base class for writing prompts."""
import re
from string import Formatter
from textwrap import dedent
from typing import Any, Callable, ClassVar, Type, TypeVar, Union

from pydantic import BaseModel

from ..enums import MessageRole
from .types import (
    AssistantMessage,
    Message,
    ModelMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)


class BasePrompt(BaseModel):
    '''The base class for working with prompts.

    This class is implemented as the base for all prompting needs across various model
    providers.

    Example:

    ```python
    from mirascope import BasePrompt


    class BookRecommendationPrompt(BasePrompt):
        """A prompt for recommending a book."""

        prompt_template = """
        SYSTEM: You are the world's greatest librarian.
        USER: Please recommend a {genre} book.
        """

        genre: str


    prompt = BookRecommendationPrompt(genre="fantasy")
    print(prompt.messages())
    #> [{"role": "user", "content": "Please recommend a fantasy book."}]

    print(prompt)
    #> Please recommend a fantasy book.
    ```
    '''

    tags: ClassVar[list[str]] = []
    prompt_template: ClassVar[str] = ""

    def __str__(self) -> str:
        """Returns the formatted template."""
        return self._format_template(self.prompt_template)

    def messages(self) -> Union[list[Message], Any]:
        """Returns the template as a formatted list of messages."""
        message_type_by_role = {
            MessageRole.SYSTEM: SystemMessage,
            MessageRole.USER: UserMessage,
            MessageRole.ASSISTANT: AssistantMessage,
            MessageRole.MODEL: ModelMessage,
            MessageRole.TOOL: ToolMessage,
        }
        return [
            message_type_by_role[MessageRole(message["role"])](
                role=message["role"], content=message["content"]
            )
            for message in self._parse_messages(list(message_type_by_role.keys()))
        ]

    def dump(
        self,
    ) -> dict[str, Any]:
        """Dumps the contents of the prompt into a dictionary."""
        return {
            "tags": self.tags,
            "template": dedent(self.prompt_template).strip("\n"),
            "inputs": self.model_dump(),
        }

    ############################## PRIVATE METHODS ###################################

    def _format_template(self, template: str):
        """Formats the given `template` with attributes matching template variables."""
        dedented_template = dedent(template).strip()
        template_vars = [
            var
            for _, var, _, _ in Formatter().parse(dedented_template)
            if var is not None
        ]
        return dedented_template.format(
            **{var: getattr(self, var) for var in template_vars}
        )

    def _parse_messages(self, roles: list[str]) -> list[Message]:
        """Returns messages parsed from the `template` ClassVar.

        Raises:
            ValueError: if the template contains an unknown role.
        """
        messages = []
        re_roles = "|".join(
            [role.upper() for role in roles] + ["MESSAGES"] + ["[A-Z]+"]
        )
        for match in re.finditer(
            rf"({re_roles}):((.|\n)+?)(?=({re_roles}):|\Z)",
            self.prompt_template,
        ):
            role = match.group(1).lower()
            if role == "messages":
                template_var = [
                    var
                    for _, var, _, _ in Formatter().parse(match.group(2))
                    if var is not None
                ][0]
                if not hasattr(self, template_var) or not isinstance(
                    getattr(self, template_var), list
                ):
                    raise ValueError(
                        f"MESSAGES keyword used with attribute `{template_var}`, which "
                        "is not a `list` of messages."
                    )
                messages += getattr(self, template_var)
            else:
                if role not in roles:
                    raise ValueError(f"Invalid role: {role}")
                content = self._format_template(match.group(2))
                messages.append({"role": role, "content": content})
        if len(messages) == 0:
            messages.append(
                {
                    "role": "user",
                    "content": self._format_template(self.prompt_template),
                }
            )
        return messages


BasePromptT = TypeVar("BasePromptT", bound=BasePrompt)


def tags(args: list[str]) -> Callable[[Type[BasePromptT]], Type[BasePromptT]]:
    '''A decorator for adding tags to a `BasePrompt`.

    Adding this decorator to a `BasePrompt` updates the `_tags` class attribute to the
    given value. This is useful for adding metadata to a `BasePrompt` that can be used
    for logging or filtering.

    Example:

    ```python
    from mirascope import BasePrompt, tags


    @tags(["book_recommendation", "entertainment"])
    class BookRecommendationPrompt(BasePrompt):
        prompt_template = """
        SYSTEM:
        You are the world's greatest librarian.

        USER:
        I've recently read this book: {book_title}.
        What should I read next?
        """

        book_title: [str]

    print(BookRecommendationPrompt.dump()["tags"])
    #> ['book_recommendation', 'entertainment']
    ```

    Returns:
        The decorated class with `tags` class attribute set.
    '''

    def tags_fn(model_class: Type[BasePromptT]) -> Type[BasePromptT]:
        """Updates the `tags` class attribute to the given value."""
        setattr(model_class, "tags", args)
        return model_class

    return tags_fn



"""A class for extracting structured information using Anthropic Claude models."""
from __future__ import annotations

import logging
from typing import Any, ClassVar, Generic, TypeVar

from ..base import BaseExtractor, ExtractedType
from .calls import AnthropicCall
from .tools import AnthropicTool
from .types import AnthropicCallParams

logger = logging.getLogger("mirascope")

T = TypeVar("T", bound=ExtractedType)


class AnthropicExtractor(BaseExtractor[AnthropicCall, AnthropicTool, T], Generic[T]):
    '''A class for extracting structured information using Anthropic Claude models.

    Example:

    ```python
    from typing import Literal, Type

    from mirascope.anthropic import AnthropicExtractor
    from pydantic import BaseModel


    class TaskDetails(BaseModel):
        title: str
        priority: Literal["low", "normal", "high"]
        due_date: str


    class TaskExtractor(AnthropicExtractor[TaskDetails]):
        extract_schema: Type[TaskDetails] = TaskDetails

        prompt_template = """
        Please extract the task details:
        {task}
        """

        task: str


    task_description = "Submit quarterly report by next Friday. Task is high priority."
    task = TaskExtractor(task=task_description).extract(retries=3)
    assert isinstance(task, TaskDetails)
    print(task)
    #> title='Submit quarterly report' priority='high' due_date='next Friday'
    ```
    '''

    call_params: ClassVar[AnthropicCallParams] = AnthropicCallParams()

    def extract(self, retries: int = 0, **kwargs: Any) -> T:
        """Extracts `extract_schema` from the Anthropic call response.

        The `extract_schema` is converted into an `AnthropicTool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of Anthropics's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
        """
        return self._extract(AnthropicCall, AnthropicTool, retries, **kwargs)

    async def extract_async(self, retries: int = 0, **kwargs: Any) -> T:
        """Asynchronously extracts `extract_schema` from the Anthropic call response.

        The `extract_schema` is converted into an `AnthropicTool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of Anthropic's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
        """
        return await self._extract_async(
            AnthropicCall, AnthropicTool, retries, **kwargs
        )



"""A module for interacting with Anthropic models."""
from .calls import AnthropicCall
from .extractors import AnthropicExtractor
from .tools import AnthropicTool
from .types import (
    AnthropicCallParams,
    AnthropicCallResponse,
    AnthropicCallResponseChunk,
)



"""The Mirascope CLI prompt management tool.

Typical usage example:

    Initialize the environment:
        $ mirascope init mirascope

    Create a prompt in the prompts directory:
        prompts/my_prompt.py
    
    Add the prompt to create a version:
        $ mirascope add my_prompt

    Iterate on the prompt in the prompts directory:

    Check the status of the prompt:
        $ mirascope status my_prompt

    Add the prompt to create a new version:
        $ mirascope add my_prompt

    Switch between prompts:
        $ mirascope use my_prompt 0001
"""


from typer import Typer

from .commands import (
    add_command,
    init_command,
    remove_command,
    status_command,
    use_command,
)

app = Typer()

app.command(name="add", help="Add a prompt")(add_command)
app.command(name="status", help="Check status of prompt(s)")(status_command)
app.command(name="use", help="Use a prompt")(use_command)
app.command(name="remove", help="Remove a prompt")(remove_command)
app.command(name="init", help="Initialize mirascope project")(init_command)



"""The status command for the Mirascope CLI."""
import os
from typing import Optional

from typer import Argument

from ..utils import (
    check_status,
    get_user_mirascope_settings,
    parse_prompt_file_name,
    prompts_directory_files,
)


def status_command(
    prompt_file_name: Optional[str] = Argument(
        help="Prompt to check status on",
        autocompletion=prompts_directory_files,
        parser=parse_prompt_file_name,
        default=None,
    ),
) -> None:
    """Checks the status of the current prompt or prompts.

    If a prompt is specified, the status of that prompt is checked. Otherwise, the
    status of all promps are checked. If a prompt has changed, the path to the prompt
    is printed.

    Args:
        prompt_file_name: (Optional) The name of the prompt file to check status on.

    Raises:
        FileNotFoundError: If the file is not found in the specified prompts directory.
    """
    mirascope_settings = get_user_mirascope_settings()
    version_directory_path = mirascope_settings.versions_location

    # If a prompt is specified, check the status of that prompt
    if prompt_file_name:
        used_prompt_path = check_status(mirascope_settings, prompt_file_name)
        if used_prompt_path:
            print(f"Prompt {used_prompt_path} has changed.")
        else:
            print("No changes detected.")
    else:  # Otherwise, check the status of all prompts
        directores_changed: list[str] = []
        for _, directories, _ in os.walk(version_directory_path):
            for directory in directories:
                used_prompt_path = check_status(mirascope_settings, directory)
                if used_prompt_path:
                    directores_changed.append(used_prompt_path)
        if len(directores_changed) > 0:
            print("The following prompts have changed:")
            for prompt in directores_changed:
                print(f"\t{prompt}".expandtabs(4))
        else:
            print("No changes detected.")



"""A module for prompting Mistral API."""
import datetime
from typing import Any, AsyncGenerator, ClassVar, Generator

from mistralai.async_client import MistralAsyncClient
from mistralai.client import MistralClient
from mistralai.constants import ENDPOINT

from ..base import BaseCall
from ..base.types import Message
from ..enums import MessageRole
from .tools import MistralTool
from .types import MistralCallParams, MistralCallResponse, MistralCallResponseChunk


class MistralCall(BaseCall[MistralCallResponse, MistralCallResponseChunk, MistralTool]):
    """A class for" prompting Mistral's chat API.

    Example:

    ```python
    from mirascope.mistral import MistralCall

    class BookRecommender(MistralCall):
        prompt_template = "Please recommend a {genre} book"

        genre: str

    response = BookRecommender(genre="fantasy").call()
    print(response.content)
    #> There are many great books to read, it ultimately depends...
    ```
    """

    call_params: ClassVar[MistralCallParams] = MistralCallParams()

    def messages(self) -> list[Message]:
        """Returns the template as a formatted list of messages."""
        return self._parse_messages(
            [MessageRole.SYSTEM, MessageRole.USER, MessageRole.ASSISTANT]
        )

    def call(self, **kwargs: Any) -> MistralCallResponse:
        """Makes a call to the model using this `MistralCall` instance.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            A `MistralCallResponse` instance.

        Raises:
            MistralException: raises any Mistral errors, see:
                https://github.com/mistralai/client-python/blob/main/src/mistralai/exceptions.py
        """
        kwargs, tool_types = self._setup(kwargs, MistralTool)
        client = MistralClient(
            api_key=self.api_key,
            endpoint=self.base_url if self.base_url else ENDPOINT,
        )
        start_time = datetime.datetime.now().timestamp() * 1000
        completion = client.chat(messages=self.messages(), **kwargs)
        return MistralCallResponse(
            response=completion,
            tool_types=tool_types,
            start_time=start_time,
            end_time=datetime.datetime.now().timestamp() * 1000,
        )

    async def call_async(self, **kwargs: Any) -> MistralCallResponse:
        """Makes an asynchronous call to the model using this `MistralCall` instance.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            A `MistralCallResponse` instance.

        Raises:
            MistralException: raises any Mistral errors, see:
                https://github.com/mistralai/client-python/blob/main/src/mistralai/exceptions.py
        """
        kwargs, tool_types = self._setup(kwargs, MistralTool)
        client = MistralAsyncClient(
            api_key=self.api_key,
            endpoint=self.base_url if self.base_url else ENDPOINT,
        )
        start_time = datetime.datetime.now().timestamp() * 1000
        completion = await client.chat(messages=self.messages(), **kwargs)
        return MistralCallResponse(
            response=completion,
            tool_types=tool_types,
            start_time=start_time,
            end_time=datetime.datetime.now().timestamp() * 1000,
        )

    def stream(self, **kwargs: Any) -> Generator[MistralCallResponseChunk, None, None]:
        """Streams the response for a call using this `MistralCall` instance.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            A `MistralCallResponseChunk` for each chunk of the response.

        Raises:
            MistralException: raises any Mistral errors, see:
                https://github.com/mistralai/client-python/blob/main/src/mistralai/exceptions.py
        """
        kwargs, tool_types = self._setup(kwargs, MistralTool)
        client = MistralClient(
            api_key=self.api_key,
            endpoint=self.base_url if self.base_url else ENDPOINT,
        )
        stream = client.chat_stream(messages=self.messages(), **kwargs)
        for chunk in stream:
            yield MistralCallResponseChunk(chunk=chunk, tool_types=tool_types)

    async def stream_async(
        self, **kwargs: Any
    ) -> AsyncGenerator[MistralCallResponseChunk, None]:
        """Streams the response for an asynchronous call using this `MistralCall`.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            A `MistralCallResponseChunk` for each chunk of the response.

        Raises:
            MistralException: raises any Mistral errors, see:
                https://github.com/mistralai/client-python/blob/main/src/mistralai/exceptions.py
        """
        kwargs, tool_types = self._setup(kwargs, MistralTool)
        client = MistralAsyncClient(
            api_key=self.api_key,
            endpoint=self.base_url if self.base_url else ENDPOINT,
        )
        stream = client.chat_stream(messages=self.messages(), **kwargs)
        async for chunk in stream:
            yield MistralCallResponseChunk(chunk=chunk, tool_types=tool_types)



"""Base types and abstract interfaces for typing LLM calls."""

from abc import ABC, abstractmethod
from inspect import isclass
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
)

from pydantic import BaseModel, ConfigDict
from typing_extensions import Required, TypedDict

from .tools import BaseTool
from .utils import convert_function_to_tool


class SystemMessage(TypedDict, total=False):
    """A message with the `system` role.

    Attributes:
        role: The role of the message's author, in this case `system`.
        content: The contents of the message.
    """

    role: Required[Literal["system"]]
    content: Required[str]


class UserMessage(TypedDict, total=False):
    """A message with the `user` role.

    Attributes:
        role: The role of the message's author, in this case `user`.
        content: The contents of the message.
    """

    role: Required[Literal["user"]]
    content: Required[str]


class AssistantMessage(TypedDict, total=False):
    """A message with the `assistant` role.

    Attributes:
        role: The role of the message's author, in this case `assistant`.
        content: The contents of the message.
    """

    role: Required[Literal["assistant"]]
    content: Required[str]


class ModelMessage(TypedDict, total=False):
    """A message with the `model` role.

    Attributes:
        role: The role of the message's author, in this case `model`.
        content: The contents of the message.
    """

    role: Required[Literal["model"]]
    content: Required[str]


class ToolMessage(TypedDict, total=False):
    """A message with the `tool` role.

    Attributes:
        role: The role of the message's author, in this case `tool`.
        content: The contents of the message.
    """

    role: Required[Literal["tool"]]
    content: Required[str]


Message = Union[SystemMessage, UserMessage, AssistantMessage, ToolMessage]


ResponseT = TypeVar("ResponseT", bound=Any)
BaseToolT = TypeVar("BaseToolT", bound=BaseTool)


class BaseCallParams(BaseModel, Generic[BaseToolT]):
    """The parameters with which to make a call."""

    model: str
    tools: Optional[list[Union[Callable, Type[BaseToolT]]]] = None

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def kwargs(
        self,
        tool_type: Optional[Type[BaseToolT]] = None,
        exclude: Optional[set[str]] = None,
    ) -> dict[str, Any]:
        """Returns all parameters for the call as a keyword arguments dictionary."""
        extra_exclude = {"tools"}
        exclude = extra_exclude if exclude is None else exclude.union(extra_exclude)
        kwargs = {
            key: value
            for key, value in self.model_dump(exclude=exclude).items()
            if value is not None
        }
        if not self.tools or tool_type is None:
            return kwargs
        kwargs["tools"] = [
            tool if isclass(tool) else convert_function_to_tool(tool, tool_type)
            for tool in self.tools
        ]
        return kwargs


class BaseCallResponse(BaseModel, Generic[ResponseT, BaseToolT], ABC):
    """A base abstract interface for LLM call responses.

    Attributes:
        response: The original response from whichever model response this wraps.
    """

    response: ResponseT
    tool_types: Optional[list[Type[BaseToolT]]] = None
    start_time: float  # The start time of the completion in ms
    end_time: float  # The end time of the completion in ms

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    @property
    @abstractmethod
    def tools(self) -> Optional[list[BaseToolT]]:
        """Returns the tools for the 0th choice message."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def tool(self) -> Optional[BaseToolT]:
        """Returns the 0th tool for the 0th choice message."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def content(self) -> str:
        """Should return the string content of the response.

        If there are multiple choices in a response, this method should select the 0th
        choice and return it's string content.

        If there is no string content (e.g. when using tools), this method must return
        the empty string.
        """
        ...  # pragma: no cover


ChunkT = TypeVar("ChunkT", bound=Any)


class BaseCallResponseChunk(BaseModel, Generic[ChunkT, BaseToolT], ABC):
    """A base abstract interface for LLM streaming response chunks.

    Attributes:
        response: The original response chunk from whichever model response this wraps.
    """

    chunk: ChunkT
    tool_types: Optional[list[Type[BaseToolT]]] = None

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    @property
    @abstractmethod
    def content(self) -> str:
        """Should return the string content of the response chunk.

        If there are multiple choices in a chunk, this method should select the 0th
        choice and return it's string content.

        If there is no string content (e.g. when using tools), this method must return
        the empty string.
        """
        ...  # pragma: no cover



"""A base interface for using tools (function calling) when calling LLMs."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, Type, TypeVar, Union

from pydantic import BaseModel, ConfigDict
from pydantic.json_schema import SkipJsonSchema

DEFAULT_TOOL_DOCSTRING = " ".join(
    """
Correctly formatted and typed parameters extracted from the completion. Must include
required parameters and may exclude optional parameters unless present in the text.
    """.strip().split("\n")
)

BaseType = Union[
    str,
    int,
    float,
    bool,
    list,
    set,
    tuple,
]

ToolCallT = TypeVar("ToolCallT", bound=Any)


class BaseTool(BaseModel, Generic[ToolCallT], ABC):
    """A base class for easy use of tools with prompts.

    `BaseTool` is an abstract class interface and should not be used directly. When
    implementing a class that extends `BaseTool`, you must include the original
    `tool_call` from which this till was instantiated. Make sure to skip `tool_call`
    when generating the schema by annotating it with `SkipJsonSchema`.
    """

    tool_call: SkipJsonSchema[ToolCallT]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def args(self) -> dict[str, Any]:
        """The arguments of the tool as a dictionary."""
        return self.model_dump(exclude={"tool_call"})

    @property
    def fn(self) -> Callable:
        """Returns the function that the tool describes."""
        raise RuntimeError("Tool does not have an attached function.")

    @classmethod
    def tool_schema(cls) -> Any:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined."""
        model_schema = cls.model_json_schema()

        fn = {
            "name": model_schema.pop("title"),
            "description": model_schema.pop("description")
            if "description" in model_schema
            else DEFAULT_TOOL_DOCSTRING,
        }
        if model_schema["properties"]:
            fn["parameters"] = model_schema

        return fn

    @classmethod
    @abstractmethod
    def from_tool_call(cls, tool_call: ToolCallT) -> BaseTool:
        """Extracts an instance of the tool constructed from a tool call response."""
        ...  # pragma: no cover

    @classmethod
    @abstractmethod
    def from_model(cls, model: Type[BaseModel]) -> Type[BaseTool]:
        """Constructs a `BaseTool` type from a `BaseModel` type."""
        ...  # pragma: no cover

    @classmethod
    @abstractmethod
    def from_fn(cls, fn: Callable) -> Type[BaseTool]:
        """Constructs a `BaseTool` type from a function."""
        ...  # pragma: no cover

    @classmethod
    @abstractmethod
    def from_base_type(cls, base_type: Type[BaseType]) -> Type[BaseTool]:
        """Constructs a `BaseTool` type from a `BaseType` type."""
        ...  # pragma: no cover



"""Prompts with WandB and OpenAI integration to support logging functionality."""
from __future__ import annotations

import datetime
from typing import Any, Generic, Literal, Optional, TypeVar, Union

from pydantic import BaseModel
from wandb.sdk.data_types.trace_tree import Trace

from ..base import BasePrompt, BaseType, ExtractedType
from ..openai import (
    OpenAICall,
    OpenAICallResponse,
    OpenAIExtractor,
    OpenAITool,
)

BaseTypeT = TypeVar("BaseTypeT", bound=BaseType)
BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


def trace(
    call: Union[WandbOpenAICall, WandbOpenAIExtractor],
    response: OpenAICallResponse,
    parent: Optional[Trace],
    **kwargs: Any,
) -> Trace:
    """Returns a trace connected to parent.

    Args:
        response: The completion to trace. Handles `OpenAIChatCompletion` output
            from both standard OpenAI chat completions, and `BaseModel` for
            extractions.
        parent: The parent trace to connect to.

    Returns:
        The created trace, connected to the parent.
    """
    tool = response.tool
    if tool is not None:
        outputs = {
            "assistant": tool.model_dump(),
            "tool_output": tool.fn(**tool.args),
        }
    else:
        outputs = {"assistant": response.content}

    metadata = {
        "call_params": call.call_params.model_copy(update=kwargs).kwargs(OpenAITool)
    }
    if response.response.usage is not None:
        metadata["usage"] = response.response.usage.model_dump()
    span = Trace(
        name=call.__class__.__name__,
        kind=call.span_type,
        status_code="success",
        status_message=None,
        metadata=metadata,
        start_time_ms=round(response.start_time),
        end_time_ms=round(response.end_time),
        inputs={message["role"]: message["content"] for message in call.messages()},
        outputs=outputs,
    )
    if parent:
        parent.add_child(span)
    return span


def trace_error(
    call: Union[WandbOpenAICall, WandbOpenAIExtractor],
    error: Exception,
    parent: Optional[Trace],
    start_time: float,
    **kwargs: Any,
) -> Trace:
    """Returns an error trace connected to parent.

    Start time is set to time of prompt creation, and end time is set to the time
    function is called.

    Args:
        error: The error to trace.
        parent: The parent trace to connect to.
        start_time: The time the call to OpenAI was started.

    Returns:
        The created error trace, connected to the parent.
    """
    span = Trace(
        name=call.__class__.__name__,
        kind=call.span_type,
        status_code="error",
        status_message=str(error),
        metadata={"call_params": call.call_params.model_copy(update=kwargs).kwargs()},
        start_time_ms=round(start_time),
        end_time_ms=round(datetime.datetime.now().timestamp() * 1000),
        inputs={message["role"]: message["content"] for message in call.messages()},
        outputs=None,
    )
    if parent:
        parent.add_child(span)
    return span


class WandbBasePrompt(BasePrompt):
    """An extension of `BasePrompt` with Weights & Biases specific functionality."""

    span_type: Literal["tool", "llm", "chain", "agent"]


class WandbOpenAICall(OpenAICall, WandbBasePrompt):
    '''A `OpenAICall` with added convenience for integrating with Weights & Biases.

    Use this class's built in `call_with_trace` method to log traces to WandB along with
    your calls to OpenAI. These calls will include all of the additional metadata
    information such as the prompt template, template variables, and more.

    Example:

    ```python
    import os

    from mirascope.wandb.openai import WandbOpenAICall
    import wandb

    wandb.login(key="YOUR_WANDB_API_KEY")
    wandb.init(project="wandb_logged_chain")

    os.environ["OPENAI_API_KEY"] = "YOUR_OPENAI_API_KEY"


    class BookRecommender(WandbOpenAICall):
        prompt_template = """
        SYSTEM:
        You are the world's greatest librarian.

        USER:
        Please recommend a {genre} book.
        """

        genre: str


    response, span = BookRecommender(span_type="llm", genre="fantasy").call_with_trace()
    #           ^ this is a `Span` returned from the trace (or trace error).
    ```
    '''

    def call_with_trace(
        self,
        parent: Optional[Trace] = None,
        **kwargs: Any,
    ) -> tuple[Optional[OpenAICallResponse], Trace]:
        """Creates an OpenAI chat completion and logs it via a W&B `Trace`.

        Args:
            parent: The parent trace to connect to.

        Returns:
            A tuple containing the completion and its trace (which has been connected
                to the parent).
        """
        try:
            start_time = datetime.datetime.now().timestamp() * 1000
            response = super().call()
            span = trace(self, response, parent, **kwargs)
            return response, span
        except Exception as e:
            return None, trace_error(self, e, parent, start_time, **kwargs)


T = TypeVar("T", bound=ExtractedType)


class WandbOpenAIExtractor(OpenAIExtractor[T], WandbBasePrompt, Generic[T]):
    '''A `OpenAIExtractor` with added convenience for integrating with Weights & Biases.

    Use this class's built in `extract_with_trace` method to log traces to WandB along
    with your calls to OpenAI. These calls will include all of the additional metadata
    information such as the prompt template, template variables, and more.

    Example:

    ```python
    import os
    from typing import Type

    from mirascope.wandb.openai import WandbOpenAIExtractor
    from pydantic import BaseModel
    import wandb

    wandb.login(key="YOUR_WANDB_API_KEY")
    wandb.init(project="wandb_logged_chain")

    os.environ["OPENAI_API_KEY"] = "YOUR_OPENAI_API_KEY"


    class Book(BaseModel):
        title: str
        author: str


    class BookRecommender(WandbOpenAIExtractor[Book]):
        extract_schema: Type[Book] = Book
        prompt_template = """
        SYSTEM:
        You are the world's greatest librarian.

        USER:
        Please recommend a {genre} book.
        """

        genre: str


    book, span = BookRecommender(span_type="tool", genre="fantasy").extract_with_trace()
    #           ^ this is a `Span` returned from the trace (or trace error).
    ```
    '''

    def extract_with_trace(
        self,
        parent: Optional[Trace] = None,
        retries: int = 0,
        **kwargs: Any,
    ) -> tuple[Optional[T], Trace]:
        """Extracts `extract_schema` from the OpenAI call response and traces it.

        The `extract_schema` is converted into an `OpenAITool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of OpenAI's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
            OpenAIError: raises any OpenAI errors, see:
                https://platform.openai.com/docs/guides/error-codes/api-errors
        """
        try:
            start_time = datetime.datetime.now().timestamp() * 1000
            model = super().extract(retries=retries, **kwargs)
            span = trace(self, model._response, parent, **kwargs)  # type: ignore
            return model, span
        except Exception as e:
            return None, trace_error(self, e, parent, start_time, **kwargs)



"""Classes for using tools with Anthropic's Claude API."""
from __future__ import annotations

import xml.etree.ElementTree as ET
from json import JSONDecodeError, loads
from textwrap import dedent
from typing import Any, Callable, Type, TypeVar, Union, cast

from pydantic import BaseModel

from ..base import BaseTool, BaseType
from ..base.utils import (
    convert_base_model_to_tool,
    convert_base_type_to_tool,
    convert_function_to_tool,
)

BaseTypeT = TypeVar("BaseTypeT", bound=BaseType)


class AnthropicTool(BaseTool[ET.Element]):
    '''A base class for easy use of tools with the Anthropic Claude client.

    `AnthropicTool` internally handles the logic that allows you to use tools with
    simple calls such as `AnthropicCallResponse.tool` or `AnthropicTool.fn`, as seen in
    the example below.

    Example:

    ```python
    from mirascope import AnthropicCall, AnthropicCallParams


    def animal_matcher(fav_food: str, fav_color: str) -> str:
        """Tells you your most likely favorite animal from personality traits.

        Args:
            fav_food: your favorite food.
            fav_color: your favorite color.

        Returns:
            The animal most likely to be your favorite based on traits.
        """
        return "Your favorite animal is the best one, a frog."


    class AnimalMatcher(AnthropicCall):
        prompt_template = """
        Tell me my favorite animal if my favorite food is {food} and my
        favorite color is {color}.
        """

        food: str
        color: str

        call_params = AnthropicCallParams(tools=[animal_matcher])


    response = AnimalMatcher(food="pizza", color="red").call
    tool = response.tool
    print(tool.fn(**tool.args))
    #> Your favorite animal is the best one, a frog.
    ```
    '''

    @classmethod
    def tool_schema(cls) -> str:
        """Constructs XML tool schema string for use with Anthropic's Claude API."""
        json_schema = super().tool_schema()
        tool_schema = (
            dedent(
                """
                <tool_description>
                <tool_name>{name}</tool_name>
                <description>
                {description}
                </description>
                """
            )
            .strip()
            .format(name=cls.__name__, description=json_schema["description"])
        )

        tool_schema += "\n"
        if "parameters" in json_schema:
            tool_schema += _process_schema(json_schema["parameters"])
        tool_schema += "</tool_description>"
        return tool_schema

    @classmethod
    def from_tool_call(cls, tool_call: ET.Element) -> AnthropicTool:
        """Extracts an instance of the tool constructed from a tool call response.

        Given the `<invoke>...</invoke>` block in a `Message` from an Anthropic call
        response, this method parses out the XML defining the tool call and creates an
        `AnthropicTool` instance from it.

        Args:
            tool_call: The XML `str` from which to extract the tool.

        Returns:
            An instance of the tool constructed from the tool call.

        Raises:
            ValidationError: if the tool call doesn't match the tool schema.
        """

        def _parse_xml_element(element: ET.Element) -> Union[str, dict, list]:
            """Recursively parse an XML element into a Python data structure."""
            children = list(element)
            if not children:
                if element.text:
                    text = element.text.strip().replace("\\n", "").replace("\\", "")
                    # Attempt to load JSON-like strings as Python data structures
                    try:
                        return loads(text)
                    except JSONDecodeError:
                        return text
                else:
                    return ""

            if (
                len(children) > 1
                and all(child.tag == children[0].tag for child in children)
            ) or (
                len(children) == 1
                and (children[0].tag == "entry" or children[0].tag == "item")
            ):
                if children[0].tag == "entry":
                    dict_result = {}
                    for child in children:
                        key, value = child.find("key"), child.find("value")
                        if key is not None and value is not None:
                            dict_result[key.text] = _parse_xml_element(value)
                    return dict_result
                else:
                    return [_parse_xml_element(child) for child in children]

            result: dict[str, Any] = {}
            for child in children:
                child_value = _parse_xml_element(child)
                result[child.tag] = child_value
            return result

        parameters = cast(dict[str, Any], _parse_xml_element(tool_call))
        parameters["tool_call"] = tool_call
        return cls.model_validate(parameters)

    @classmethod
    def from_model(cls, model: Type[BaseModel]) -> Type[AnthropicTool]:
        """Constructs a `AnthropicTool` type from a `BaseModel` type."""
        return convert_base_model_to_tool(model, AnthropicTool)

    @classmethod
    def from_fn(cls, fn: Callable) -> Type[AnthropicTool]:
        """Constructs a `AnthropicTool` type from a function."""
        return convert_function_to_tool(fn, AnthropicTool)

    @classmethod
    def from_base_type(cls, base_type: Type[BaseTypeT]) -> Type[AnthropicTool]:
        """Constructs a `AnthropicTool` type from a `BaseType` type."""
        return convert_base_type_to_tool(base_type, AnthropicTool)


def _process_schema(schema: dict[str, Any]) -> str:
    schema_xml = ""
    if "$defs" in schema:
        schema_xml += "<definitions>\n"
        for def_name, definition in schema["$defs"].items():
            schema_xml += f"<definition name='{def_name}'>\n"
            schema_xml += _process_property(definition)
            schema_xml += "</definition>\n"
        schema_xml += "</definitions>"
    if "properties" in schema:
        schema_xml += "<parameters>\n"
        for prop, definition in schema["properties"].items():
            schema_xml += "<parameter>\n"
            schema_xml += f"<name>{prop}</name>\n"
            schema_xml += _process_property(definition)
            schema_xml += "</parameter>\n"
        schema_xml += "</parameters>\n"
    return schema_xml


# TODO: consider using ET subelements instead of string concat
def _process_property(definition: dict[str, Any]) -> str:
    prop_xml = ""
    prop_type = definition.get("type", "object")

    if "$ref" in definition:
        ref_name = definition["$ref"].split("/")[-1]
        prop_xml += f"<reference name='{ref_name}'/>\n"
    elif prop_type == "array":
        prop_xml += "<type>list</type>\n"
        items_def = None
        if "items" in definition:
            items_def = definition["items"]
        elif "prefixItems" in definition:
            items_def = definition["prefixItems"]
        if items_def:
            if isinstance(items_def, dict):
                prop_xml += "<element_type>\n"
                prop_xml += _process_property(items_def)
                prop_xml += "</element_type>\n"
            else:
                for item in items_def:
                    prop_xml += "<element_type>\n"
                    prop_xml += _process_property(item)
                    prop_xml += "</element_type>\n"
        if "maxItems" in definition:
            prop_xml += f"<maxItems>{definition['maxItems']}</maxItems>\n"
        if "minItems" in definition:
            prop_xml += f"<minItems>{definition['minItems']}</minItems>\n"
        if "uniqueItems" in definition:
            prop_xml += (
                f"<uniqueItems>{str(definition['uniqueItems']).lower()}</uniqueItems>\n"
            )
    elif "enum" in definition:
        prop_xml += "<type>enum</type>\n"
        prop_xml += "<values>\n"
        for value in definition["enum"]:
            prop_xml += f"<value>{value}</value>\n"
        prop_xml += "</values>\n"
        prop_xml += f"<value_type>{prop_type}</value_type>\n"
    elif "anyOf" in definition:
        prop_xml += "<type>union</type>\n"
        prop_xml += "<options>\n"
        for option in definition["anyOf"]:
            prop_xml += "<option>\n"
            prop_xml += _process_property(option)
            prop_xml += "</option>\n"
        prop_xml += "</options>\n"
    elif "const" in definition:
        prop_xml += "<type>literal</type>\n"
        prop_xml += f"<value>{definition['const']}</value>\n"
    elif "additionalProperties" in definition:
        prop_xml += "<type>dictionary</type>\n"
        prop_xml += "<entry>\n"
        prop_xml += "<key>\n<type>string</type>\n</key>\n"
        prop_xml += "<value>\n"
        prop_xml += _process_property(definition["additionalProperties"])
        prop_xml += "</value>\n"
        prop_xml += "</entry>\n"
    elif prop_type in ["string", "number", "integer", "boolean"]:
        prop_xml += f"<type>{prop_type}</type>\n"
    elif prop_type == "null":
        prop_xml += "<type>null</type>\n"
    else:
        prop_xml += "<type>object</type>\n"
        prop_xml += _process_schema(definition)

    if "description" in definition:
        prop_xml += f"<description>{definition['description']}</description>\n"
    if "default" in definition:
        default_value = definition["default"]
        if default_value is None:
            prop_xml += "<default>null</default>\n"
        else:
            prop_xml += f"<default>{default_value}</default>\n"

    return prop_xml



"""Types for working with Mistral prompts."""
from typing import Any, Optional

from mistralai.models.chat_completion import (
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    ChatCompletionResponseStreamChoice,
    ChatCompletionStreamResponse,
    ChatMessage,
    DeltaMessage,
    ToolCall,
    ToolChoice,
)
from pydantic import ConfigDict

from ..base import BaseCallParams, BaseCallResponse, BaseCallResponseChunk
from .tools import MistralTool


class MistralCallParams(BaseCallParams[MistralTool]):
    """The parameters to use when calling the Mistral API."""

    model: str = "open-mixtral-8x7b"
    endpoint: Optional[str] = None
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None
    top_p: Optional[float] = None
    random_seed: Optional[int] = None
    safe_mode: Optional[bool] = None
    safe_prompt: Optional[bool] = None
    tool_choice: Optional[ToolChoice] = None
    model_config = ConfigDict(arbitrary_types_allowed=True)


class MistralCallResponse(BaseCallResponse[ChatCompletionResponse, MistralTool]):
    """Convenience wrapper for Mistral's chat model completions.

    When using Mirascope's convenience wrappers to interact with Mistral models via
    `MistralCall`, responses using `MistralCall.call()` will return a
    `MistralCallResponse`, whereby the implemented properties allow for simpler syntax
    and a convenient developer experience.

    Example:

    ```python
    from mirascope.mistral import MistralCall

    class BookRecommender(MistralCall):
        prompt_template = "Please recommend a {genre} book"

        genre: str

    response = Bookrecommender(genre="fantasy").call()
    print(response.content)
    #> The Name of the Wind

    print(response.message)
    #> ChatMessage(content='The Name of the Wind', role='assistant',
    #  function_call=None, tool_calls=None)

    print(response.choices)
    #> [Choice(finish_reason='stop', index=0, logprobs=None,
    #  message=ChatMessage(content='The Name of the Wind', role='assistant',
    #  function_call=None, tool_calls=None))]
    ```

    """

    @property
    def choices(self) -> list[ChatCompletionResponseChoice]:
        """Returns the array of chat completion choices."""
        return self.response.choices

    @property
    def choice(self) -> ChatCompletionResponseChoice:
        """Returns the 0th choice."""
        return self.choices[0]

    @property
    def message(self) -> ChatMessage:
        """Returns the message of the chat completion for the 0th choice."""
        return self.choice.message

    @property
    def content(self) -> str:
        """The content of the chat completion for the 0th choice."""
        content = self.message.content
        # We haven't seen the `list[str]` response type in practice, so for now we
        # return the first item in the list
        return content if isinstance(content, str) else content[0]

    @property
    def tool_calls(self) -> Optional[list[ToolCall]]:
        """Returns the tool calls for the 0th choice message."""
        return self.message.tool_calls

    @property
    def tools(self) -> Optional[list[MistralTool]]:
        """Returns the tools for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if not self.tool_types or not self.tool_calls or len(self.tool_calls) == 0:
            return None

        extracted_tools = []
        for tool_call in self.tool_calls:
            for tool_type in self.tool_types:
                if tool_call.function.name == tool_type.__name__:
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @property
    def tool(self) -> Optional[MistralTool]:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if not self.tool_types or not self.tool_calls or len(self.tool_calls) == 0:
            return None

        tool_call = self.tool_calls[0]
        for tool_type in self.tool_types:
            if self.tool_calls[0].function.name == tool_type.__name__:
                return tool_type.from_tool_call(tool_call)

        return None

    def dump(self) -> dict[str, Any]:
        """Dumps the response to a dictionary."""
        return {
            "start_time": self.start_time,
            "end_time": self.end_time,
            "output": self.response.model_dump(),
        }


class MistralCallResponseChunk(
    BaseCallResponseChunk[ChatCompletionStreamResponse, MistralTool]
):
    """Convenience wrapper around chat completion streaming chunks.

    When using Mirascope's convenience wrappers to interact with Mistral models via
    `MistralCall.stream`, responses will return an `MistralCallResponseChunk`, whereby
    the implemented properties allow for simpler syntax and a convenient developer
    experience.

    Example:

    ```python
    from mirascope.mistral import MistralCall


    class Math(MistralCall):
        prompt_template = "What is 1 + 2?"


    for chunk in MistralCall().stream():
        print(chunk.content)

    #> 1
    #  +
    #  2
    #   equals
    #
    #  3
    #  .
    ```
    """

    @property
    def choices(self) -> list[ChatCompletionResponseStreamChoice]:
        """Returns the array of chat completion choices."""
        return self.chunk.choices

    @property
    def choice(self) -> ChatCompletionResponseStreamChoice:
        """Returns the 0th choice."""
        return self.choices[0]

    @property
    def delta(self) -> DeltaMessage:
        """Returns the delta of the 0th choice."""
        return self.choice.delta

    @property
    def content(self) -> str:
        """Returns the content of the delta."""
        return self.delta.content if self.delta.content is not None else ""

    @property
    def tool_calls(self) -> Optional[list[ToolCall]]:
        """Returns the partial tool calls for the 0th choice message."""
        return self.delta.tool_calls



"""A module for calling Google's Gemini Chat API."""
import datetime
import logging
from typing import (
    Any,
    AsyncGenerator,
    ClassVar,
    Generator,
)

from google.generativeai import GenerativeModel  # type: ignore
from google.generativeai.types import ContentsType  # type: ignore

from ..base import BaseCall
from ..enums import MessageRole
from .tools import GeminiTool
from .types import (
    GeminiCallParams,
    GeminiCallResponse,
    GeminiCallResponseChunk,
)

logger = logging.getLogger("mirascope")


class GeminiCall(BaseCall[GeminiCallResponse, GeminiCallResponseChunk, GeminiTool]):
    '''A class for prompting Google's Gemini Chat API.

    This prompt supports the message types: USER, MODEL, TOOL

    Example:

    ```python
    from google.generativeai import configure  # type: ignore
    from mirascope.gemini import GeminiPrompt

    configure(api_key="YOUR_API_KEY")


    class BookRecommender(GeminiPrompt):
        prompt_template = """
        USER: You're the world's greatest librarian.
        MODEL: Ok, I understand I'm the world's greatest librarian. How can I help?
        USER: Please recommend some {genre} books.

        genre: str


    response = BookRecommender(genre="fantasy").call()
    print(response.call())
    #> As the world's greatest librarian, I am delighted to recommend...
    ```
    '''

    call_params: ClassVar[GeminiCallParams] = GeminiCallParams()

    def messages(self) -> ContentsType:
        """Returns the `ContentsType` messages for Gemini `generate_content`.

        Raises:
            ValueError: if the docstring contains an unknown role.
        """
        return [
            {"role": message["role"], "parts": [message["content"]]}
            for message in self._parse_messages(
                [MessageRole.MODEL, MessageRole.USER, MessageRole.TOOL]
            )
        ]

    def call(self, **kwargs: Any) -> GeminiCallResponse:
        """Makes an call to the model using this `GeminiCall` instance.

        Args:
            **kwargs: Additional keyword arguments that will be used for generating the
                response. These will override any existing argument settings in call
                params.

        Returns:
            A `GeminiCallResponse` instance.
        """
        kwargs, tool_types = self._setup(kwargs, GeminiTool)
        gemini_pro_model = GenerativeModel(model_name=kwargs.pop("model"))
        start_time = datetime.datetime.now().timestamp() * 1000
        response = gemini_pro_model.generate_content(
            self.messages(),
            stream=False,
            tools=kwargs.pop("tools") if "tools" in kwargs else None,
            **kwargs,
        )
        return GeminiCallResponse(
            response=response,
            tool_types=tool_types,
            start_time=start_time,
            end_time=datetime.datetime.now().timestamp() * 1000,
        )

    async def call_async(self, **kwargs: Any) -> GeminiCallResponse:
        """Makes an asynchronous call to the model using this `GeminiCall` instance.

        Args:
            **kwargs: Additional keyword arguments that will be used for generating the
                response. These will override any existing argument settings in call
                params.

        Returns:
            A `GeminiCallResponse` instance.
        """
        kwargs, tool_types = self._setup(kwargs, GeminiTool)
        gemini_pro_model = GenerativeModel(model_name=kwargs.pop("model"))
        start_time = datetime.datetime.now().timestamp() * 1000
        response = await gemini_pro_model.generate_content_async(
            self.messages(),
            stream=False,
            tools=kwargs.pop("tools") if "tools" in kwargs else None,
            **kwargs,
        )
        return GeminiCallResponse(
            response=response,
            tool_types=tool_types,
            start_time=start_time,
            end_time=datetime.datetime.now().timestamp() * 1000,
        )

    def stream(self, **kwargs: Any) -> Generator[GeminiCallResponseChunk, None, None]:
        """Streams the response for a call using this `GeminiCall`.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Yields:
            A `GeminiCallResponseChunk` for each chunk of the response.
        """
        kwargs, tool_types = self._setup(kwargs, GeminiTool)
        gemini_pro_model = GenerativeModel(model_name=kwargs.pop("model"))
        stream = gemini_pro_model.generate_content(
            self.messages(),
            stream=True,
            tools=kwargs.pop("tools") if "tools" in kwargs else None,
            **kwargs,
        )
        for chunk in stream:
            yield GeminiCallResponseChunk(chunk=chunk, tool_types=tool_types)

    async def stream_async(
        self, **kwargs: Any
    ) -> AsyncGenerator[GeminiCallResponseChunk, None]:
        """Streams the response asynchronously for a call using this `GeminiCall`.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Yields:
            A `GeminiCallResponseChunk` for each chunk of the response.
        """
        kwargs, tool_types = self._setup(kwargs, GeminiTool)
        gemini_pro_model = GenerativeModel(model_name=kwargs.pop("model"))
        stream = await gemini_pro_model.generate_content_async(
            self.messages(),
            stream=True,
            tools=kwargs.pop("tools") if "tools" in kwargs else None,
            **kwargs,
        )
        async for chunk in stream:
            yield GeminiCallResponseChunk(chunk=chunk, tool_types=tool_types)



"""Types for interacting with Google's Gemini models using Mirascope."""
from typing import Any, Optional, TypeVar

from google.generativeai.types import GenerateContentResponse  # type: ignore

from ..base import BaseCallParams, BaseCallResponse, BaseCallResponseChunk, BaseTool
from .tools import GeminiTool

BaseToolT = TypeVar("BaseToolT", bound=BaseTool)


class GeminiCallParams(BaseCallParams[GeminiTool]):
    """The parameters to use when calling the Gemini API calls.

    Example:

    ```python
    from mirascope.gemini import GeminiCall, GeminiCallParams


    class BookRecommendation(GeminiPrompt):
        prompt_template = "Please recommend a {genre} book"

        genre: str

        call_params = GeminiCallParams(
            model="gemini-1.0-pro-001",
            generation_config={"candidate_count": 2},
        )


    response = BookRecommender(genre="fantasy").call()
    print(response.content)
    #> The Name of the Wind
    ```
    """

    model: str = "gemini-1.0-pro"
    generation_config: Optional[dict[str, Any]] = {"candidate_count": 1}
    safety_settings: Optional[Any] = None
    request_options: Optional[dict[str, Any]] = None


class GeminiCallResponse(BaseCallResponse[GenerateContentResponse, GeminiTool]):
    """Convenience wrapper around Gemini's `GenerateContentResponse`.

    When using Mirascope's convenience wrappers to interact with Gemini models via
    `GeminiCall`, responses using `GeminiCall.call()` will return a
    `GeminiCallResponse`, whereby the implemented properties allow for simpler syntax
    and a convenient developer experience.

    Example:

    ```python
    from mirascope.gemini import GeminiPrompt


    class BookRecommender(GeminiPrompt):
        prompt_template = "Please recommend a {genre} book"

        genre: str


    response = BookRecommender(genre="fantasy").call()
    print(response.content)
    #> The Lord of the Rings
    ```
    """

    @property
    def tools(self) -> Optional[list[GeminiTool]]:
        """Returns the list of tools for the 0th candidate's 0th content part."""
        if self.tool_types is None:
            return None

        tool_calls = [
            part.function_call for part in self.response.candidates[0].content.parts
        ]

        extracted_tools = []
        for tool_call in tool_calls:
            for tool_type in self.tool_types:
                if tool_call.name == tool_type.__name__:
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @property
    def tool(self) -> Optional[GeminiTool]:
        """Returns the 0th tool for the 0th candidate's 0th content part.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if self.tool_types is None:
            return None

        tool_call = self.response.candidates[0].content.parts[0].function_call
        for tool_type in self.tool_types:
            if tool_call.name == tool_type.__name__:
                return tool_type.from_tool_call(tool_call)

        return None

    @property
    def content(self) -> str:
        """Returns the contained string content for the 0th choice."""
        return self.response.candidates[0].content.parts[0].text

    def dump(self) -> dict[str, Any]:
        """Dumps the response to a dictionary."""
        return {
            "start_time": self.start_time,
            "end_time": self.end_time,
            "output": str(self.response),
        }


class GeminiCallResponseChunk(
    BaseCallResponseChunk[GenerateContentResponse, GeminiTool]
):
    """Convenience wrapper around chat completion streaming chunks.

    When using Mirascope's convenience wrappers to interact with Gemini models via
    `GeminiCall`, responses using `GeminiCall.stream()` will return a
    `GeminiCallResponseChunk`, whereby the implemented properties allow for simpler
    syntax and a convenient developer experience.

    Example:

    ```python
    from mirascope.gemini import GeminiCall


    class BookRecommender(GeminiCall):
        prompt_template = "Please recommend a {genre} book"

        genre: str


    for chunk in BookRecommender(genre="science fiction").stream():
        print(chunk)

    #> D
    #  u
    #
    #  ne
    #
    #  by F
    #  r
    #  an
    #  k
    #  .
    ```
    """

    @property
    def content(self) -> str:
        """Returns the chunk content for the 0th choice."""
        return self.chunk.candidates[0].content.parts[0].text



"""Classes for using tools with Google's Gemini API."""
from __future__ import annotations

from typing import Callable, Type, TypeVar

from google.ai.generativelanguage import FunctionCall
from google.generativeai.types import (  # type: ignore
    FunctionDeclaration,
    Tool,
)
from pydantic import BaseModel, ConfigDict

from ..base import (
    BaseTool,
    BaseType,
    convert_base_model_to_tool,
    convert_base_type_to_tool,
    convert_function_to_tool,
)

BaseTypeT = TypeVar("BaseTypeT", bound=BaseType)


class GeminiTool(BaseTool[FunctionCall]):
    '''A base class for easy use of tools with the Gemini API.

    `GeminiTool` internally handles the logic that allows you to use tools with simple
    calls such as `GeminiCompletion.tool` or `GeminiTool.fn`, as seen in the
    examples below.

    Example:

    ```python
    from mirascope.gemini import GeminiCall, GeminiCallParams, GeminiTool


    class CurrentWeather(GeminiTool):
        """A tool for getting the current weather in a location."""

        location: str


    class WeatherForecast(GeminiPrompt):
        prompt_template = "What is the current weather in {city}?"

        city: str

        call_params = GeminiCallParams(
            model="gemini-pro",
            tools=[CurrentWeather],
        )


    prompt = WeatherPrompt()
    forecast = WeatherForecast(city="Tokyo").call().tool
    print(forecast.location)
    #> Tokyo
    ```
    '''

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def tool_schema(cls) -> Tool:
        """Constructs a tool schema for use with the Gemini API.

        A Mirascope `GeminiTool` is deconstructed into a `Tool` schema for use with the
        Gemini API.

        Returns:
            The constructed `Tool` schema.
        """
        tool_schema = super().tool_schema()
        if "parameters" in tool_schema:
            if "$defs" in tool_schema["parameters"]:
                raise ValueError(
                    "Unfortunately Google's Gemini API cannot handle nested structures "
                    "with $defs."
                )
            tool_schema["parameters"]["properties"] = {
                prop: {
                    key: value for key, value in prop_schema.items() if key != "title"
                }
                for prop, prop_schema in tool_schema["parameters"]["properties"].items()
            }
        return Tool(function_declarations=[FunctionDeclaration(**tool_schema)])

    @classmethod
    def from_tool_call(cls, tool_call: FunctionCall) -> GeminiTool:
        """Extracts an instance of the tool constructed from a tool call response.

        Given a `GenerateContentResponse` from a Gemini chat completion response, this
        method extracts the tool call and constructs an instance of the tool.

        Args:
            tool_call: The `GenerateContentResponse` from which to extract the tool.

        Returns:
            An instance of the tool constructed from the tool call.

        Raises:
            ValueError: if the tool call doesn't have any arguments.
            ValidationError: if the tool call doesn't match the tool schema.
        """
        if not tool_call.args:
            raise ValueError("Tool call doesn't have any arguments.")
        model_json = {key: value for key, value in tool_call.args.items()}
        model_json["tool_call"] = tool_call
        return cls.model_validate(model_json)

    @classmethod
    def from_model(cls, model: Type[BaseModel]) -> Type[GeminiTool]:
        """Constructs a `GeminiTool` type from a `BaseModel` type."""
        return convert_base_model_to_tool(model, GeminiTool)

    @classmethod
    def from_fn(cls, fn: Callable) -> Type[GeminiTool]:
        """Constructs a `GeminiTool` type from a function."""
        return convert_function_to_tool(fn, GeminiTool)

    @classmethod
    def from_base_type(cls, base_type: Type[BaseTypeT]) -> Type[GeminiTool]:
        """Constructs a `GeminiTool` type from a `BaseType` type."""
        return convert_base_type_to_tool(base_type, GeminiTool)



"""Base modules for the Mirascope library."""
from .calls import BaseCall
from .extractors import BaseExtractor, ExtractedType
from .prompts import BasePrompt, tags
from .tools import BaseTool, BaseType
from .types import (
    BaseCallParams,
    BaseCallResponse,
    BaseCallResponseChunk,
    Message,
)
from .utils import (
    convert_base_model_to_tool,
    convert_base_type_to_tool,
    convert_function_to_tool,
    tool_fn,
)



"""Enum Classes for mirascope."""
from enum import Enum, EnumMeta
from typing import Any


class _Metaclass(EnumMeta):
    """Base `EnumMeta` subclass for accessing enum members directly."""

    def __getattribute__(cls, __name: str) -> Any:
        value = super().__getattribute__(__name)
        if isinstance(value, Enum):
            value = value.value
        return value


class _Enum(str, Enum, metaclass=_Metaclass):
    """Base Enum Class."""


class MirascopeCommand(_Enum):
    """CLI commands to be executed.

    - ADD: save a modified prompt to the `versions/` folder.
    - USE: load a specific version of the prompt from `versions/` as the current prompt.
    - STATUS: display if any changes have been made to prompts, and if a prompt is
        specified, displays changes for only said prompt.
    - INIT: initializes the necessary folders for prompt versioning with CLI.
    """

    ADD = "add"
    USE = "use"
    STATUS = "status"
    INIT = "init"


class MessageRole(_Enum):
    """Roles that the `BasePrompt` messages parser can parse from the template.

    SYSTEM: A system message.
    USER: A user message.
    ASSISTANT: A message response from the assistant or chat client.
    MODEL: A message response from the assistant or chat client. Model is used by
        Google's Gemini instead of assistant, which doesn't have system messages.
    TOOL: A message representing the output of calling a tool.
    """

    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    MODEL = "model"
    TOOL = "tool"



import logging
from typing import Any, ClassVar, Generic, TypeVar

from ..base import BaseExtractor, ExtractedType
from .calls import GeminiCall
from .tools import GeminiTool
from .types import GeminiCallParams

logger = logging.getLogger("mirascope")

T = TypeVar("T", bound=ExtractedType)


class GeminiExtractor(BaseExtractor[GeminiCall, GeminiTool, T], Generic[T]):
    '''A class for extracting structured information using Google's Gemini Chat models.

    Example:

    ```python
    from typing import Literal, Type
    from pydantic import BaseModel
    from mirascope.gemini import GeminiExtractor

    class TaskDetails(BaseModel):
        title: str
        priority: Literal["low", "normal", "high"]
        due_date: str

    class TaskExtractor(GeminiExtractor[TaskDetails]):
        extract_schema: Type[TaskDetails] = TaskDetails

        prompt_template = """
        USER: I need to extract task details.
        MODEL: Sure, please provide the task description.
        USER: {task}
        """

        task: str

    task_description = "Prepare the budget report by next Monday. It's a high priority task."
    task = TaskExtractor(task=task_description).extract(retries=3)
    assert isinstance(task, TaskDetails)
    print(task)
    #> title='Prepare the budget report' priority='high' due_date='next Monday'
    ```
    '''

    call_params: ClassVar[GeminiCallParams] = GeminiCallParams()

    def extract(self, retries: int = 0, **kwargs: Any) -> T:
        """Extracts `extract_schema` from the Gemini call response.

        The `extract_schema` is converted into a `GeminiTool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of Gemini's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
            GeminiError: raises any Gemini errors.
        """
        return self._extract(GeminiCall, GeminiTool, retries, **kwargs)

    async def extract_async(self, retries: int = 0, **kwargs: Any) -> T:
        """Asynchronously extracts `extract_schema` from the Gemini call response.

        The `extract_schema` is converted into a `GeminiTool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of Gemini's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
            GeminiError: raises any Gemini errors.
        """
        return await self._extract_async(GeminiCall, GeminiTool, retries, **kwargs)



"""A class for extracting structured information using OpenAI chat models."""
from __future__ import annotations

import logging
from typing import Any, ClassVar, Generic, TypeVar

from ..base import BaseExtractor, ExtractedType
from .calls import OpenAICall
from .tools import OpenAITool
from .types import OpenAICallParams

logger = logging.getLogger("mirascope")

T = TypeVar("T", bound=ExtractedType)


class OpenAIExtractor(BaseExtractor[OpenAICall, OpenAITool, T], Generic[T]):
    '''A class for extracting structured information using OpenAI chat models.

    Example:

    ```python
    from typing import Literal, Type

    from mirascope.openai import OpenAIExtractor
    from pydantic import BaseModel


    class TaskDetails(BaseModel):
        title: str
        priority: Literal["low", "normal", "high"]
        due_date: str


    class TaskExtractor(OpenAIExtractor[TaskDetails]):
        extract_schema: Type[TaskDetails] = TaskDetails

        prompt_template = """
        Please extract the task details:
        {task}
        """

        task: str


    task_description = "Submit quarterly report by next Friday. Task is high priority."
    task = TaskExtractor(task=task_description).extract(retries=3)
    assert isinstance(task, TaskDetails)
    print(task)
    #> title='Submit quarterly report' priority='high' due_date='next Friday'
    ```
    '''

    call_params: ClassVar[OpenAICallParams] = OpenAICallParams()

    def extract(self, retries: int = 0, **kwargs: Any) -> T:
        """Extracts `extract_schema` from the OpenAI call response.

        The `extract_schema` is converted into an `OpenAITool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of OpenAI's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
            OpenAIError: raises any OpenAI errors, see:
                https://platform.openai.com/docs/guides/error-codes/api-errors
        """
        return self._extract(OpenAICall, OpenAITool, retries, **kwargs)

    async def extract_async(self, retries: int = 0, **kwargs: Any) -> T:
        """Asynchronously extracts `extract_schema` from the OpenAI call response.

        The `extract_schema` is converted into an `OpenAITool`, complete with a
        description of the tool, all of the fields, and their types. This allows us to
        take advantage of OpenAI's tool/function calling functionality to extract
        information from a prompt according to the context provided by the `BaseModel`
        schema.

        Args:
            retries: The maximum number of times to retry the query on validation error.
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            The `Schema` instance extracted from the completion.

        Raises:
            AttributeError: if there is no tool in the call creation.
            ValidationError: if the schema cannot be instantiated from the completion.
            OpenAIError: raises any OpenAI errors, see:
                https://platform.openai.com/docs/guides/error-codes/api-errors
        """
        return await self._extract_async(OpenAICall, OpenAITool, retries, **kwargs)



"""A module for calling Anthropic's Claude API."""
import datetime
from textwrap import dedent
from typing import Any, AsyncGenerator, ClassVar, Generator, Optional, Type

from anthropic import Anthropic, AsyncAnthropic
from anthropic.types import MessageParam

from ..base import BaseCall
from ..enums import MessageRole
from .tools import AnthropicTool
from .types import (
    AnthropicCallParams,
    AnthropicCallResponse,
    AnthropicCallResponseChunk,
)


class AnthropicCall(BaseCall[AnthropicCallResponse, AnthropicCallResponseChunk, Any]):
    """A base class for calling Anthropic's Claude models.

    Example:

    ```python
    from mirascope.anthropic import AnthropicCall


    class BookRecommender(AnthropicCall):
        prompt_template = "Please recommend a {genre} book."

        genre: str


    response = BookRecommender(genre="fantasy").call()
    print(response.content)
    #> There are many great books to read, it ultimately depends...
    ```
    """

    call_params: ClassVar[AnthropicCallParams] = AnthropicCallParams()

    def messages(self) -> list[MessageParam]:
        """Returns the template as a formatted list of messages."""
        return self._parse_messages(
            [MessageRole.SYSTEM, MessageRole.USER, MessageRole.ASSISTANT]
        )  # type: ignore

    def call(self, **kwargs: Any) -> AnthropicCallResponse:
        """Makes a call to the model using this `AnthropicCall` instance.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            A `AnthropicCallResponse` instance.
        """
        messages, kwargs, tool_types = self._setup_anthropic_kwargs(kwargs)
        client = Anthropic(api_key=self.api_key, base_url=self.base_url)
        if self.call_params.wrapper is not None:
            client = self.call_params.wrapper(client)
        start_time = datetime.datetime.now().timestamp() * 1000
        message = client.messages.create(
            messages=messages,
            stream=False,
            **kwargs,
        )
        return AnthropicCallResponse(
            response=message,
            tool_types=tool_types,
            start_time=start_time,
            end_time=datetime.datetime.now().timestamp() * 1000,
        )

    async def call_async(self, **kwargs: Any) -> AnthropicCallResponse:
        """Makes an asynchronous call to the model using this `AnthropicCall` instance.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Returns:
            A `AnthropicCallResponse` instance.
        """
        messages, kwargs, tool_types = self._setup_anthropic_kwargs(kwargs)
        client = AsyncAnthropic(api_key=self.api_key, base_url=self.base_url)
        if self.call_params.wrapper_async is not None:
            client = self.call_params.wrapper_async(client)
        start_time = datetime.datetime.now().timestamp() * 1000
        message = await client.messages.create(
            messages=messages,
            stream=False,
            **kwargs,
        )
        return AnthropicCallResponse(
            response=message,
            tool_types=tool_types,
            start_time=start_time,
            end_time=datetime.datetime.now().timestamp() * 1000,
        )

    def stream(
        self, **kwargs: Any
    ) -> Generator[AnthropicCallResponseChunk, None, None]:
        """Streams the response for a call using this `AnthropicCall`.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Yields:
            An `AnthropicCallResponseChunk` for each chunk of the response.
        """
        messages, kwargs, tool_types = self._setup_anthropic_kwargs(kwargs)
        client = Anthropic(api_key=self.api_key, base_url=self.base_url)
        if self.call_params.wrapper is not None:
            client = self.call_params.wrapper(client)
        with client.messages.stream(messages=messages, **kwargs) as stream:
            for chunk in stream:
                yield AnthropicCallResponseChunk(chunk=chunk, tool_types=tool_types)

    async def stream_async(
        self, **kwargs: Any
    ) -> AsyncGenerator[AnthropicCallResponseChunk, None]:
        """Streams the response for an asynchronous call using this `AnthropicCall`.

        Args:
            **kwargs: Additional keyword arguments parameters to pass to the call. These
                will override any existing arguments in `call_params`.

        Yields:
            An `AnthropicCallResponseChunk` for each chunk of the response.
        """
        messages, kwargs, tool_types = self._setup_anthropic_kwargs(kwargs)
        client = AsyncAnthropic(api_key=self.api_key, base_url=self.base_url)
        if self.call_params.wrapper_async is not None:
            client = self.call_params.wrapper_async(client)
        async with client.messages.stream(messages=messages, **kwargs) as stream:
            async for chunk in stream:
                yield AnthropicCallResponseChunk(chunk=chunk, tool_types=tool_types)

    ############################## PRIVATE METHODS ###################################

    def _setup_anthropic_kwargs(
        self,
        kwargs: dict[str, Any],
    ) -> tuple[
        list[MessageParam],
        dict[str, Any],
        Optional[list[Type[AnthropicTool]]],
    ]:
        """Overrides the `BaseCall._setup` for Anthropic specific setup."""
        kwargs, tool_types = self._setup(kwargs, AnthropicTool)
        messages = self.messages()
        system_message = ""
        if messages[0]["role"] == "system":
            system_message += messages.pop(0)["content"]
        if tool_types:
            tool_schemas = kwargs.pop("tools")
            system_message += self._write_tools_system_message(tool_schemas)
            kwargs["stop_sequences"] = ["</function_calls>"]
        if system_message:
            kwargs["system"] = system_message
        return messages, kwargs, tool_types

    def _write_tools_system_message(self, tool_schemas: list[str]) -> str:
        """Returns the Anthropic Tools System Message from their guide."""
        return dedent(
            """
        In this environment you have access to a set of tools you can use to answer the user's question.

        You may call them like this:
        <function_calls>
        <invoke>
        <tool_name>$TOOL_NAME</tool_name>
        <parameters>
        <$PARAMETER_NAME>$PARAMETER_VALUE</$PARAMETER_NAME>
        ...
        </parameters>
        </invoke>
        ...
        </function_calls>

        Make sure to include all parameters in the tool schema when requested.
        If you want to call multiple tools, you should put all of the tools inside of the <function_calls> tag as multiple <invoke> elements.

        To output nested structured data, encode it as valid XML with tags and values. For example:

        List[int]:
        <parameterName>
            <item>1</item>
            <item>2</item>
            <item>3</item>
        </parameterName>

        List[object]:
        <parameterName>
            <item>
                <objectName>
                    <objectValue>value</objectValue>
                </objectName>
            </item>
        </parameterName>

        Dictionary:
        <parameterName>
            <entry>
                <key>key1</key>
                <value>value1</value>
            </entry>
            <entry>
                <key>key2</key>
                <value>value2</value>
            </entry>
        </parameterName>

        Remember, the above are just examples.
        Make sure to properly nest by wrapping elements in lists with the <item> tag and dictionary elements with <entry> as necessary.
        DO NOT FORGET THESE TAGS. Without these tags, we cannot properly parse the information you send.

        Here are the tools available:
        <tools>
        {tool_schemas}
        </tools>
            """.format(tool_schemas=tool_schemas)
        )



"""Utility functions for the mirascope library."""
from __future__ import annotations

import ast
import glob
import json
import os
import subprocess
import sys
from configparser import ConfigParser, MissingSectionHeaderError
from pathlib import Path
from typing import Any, Literal, Optional, Union

from jinja2 import Environment, FileSystemLoader

from ..enums import MirascopeCommand
from .constants import CURRENT_REVISION_KEY, LATEST_REVISION_KEY
from .schemas import (
    ClassInfo,
    FunctionInfo,
    MirascopeCliVariables,
    MirascopeSettings,
    VersionTextFile,
)

ignore_variables = {"prev_revision_id", "revision_id"}
mirascope_prompt_bases = (
    "BasePrompt",
    "OpenAICall",
    "GeminiCall",
    "AnthropicCall",
    "WandbOpenAICall",
)


class PromptAnalyzer(ast.NodeVisitor):
    """Utility class for analyzing a Mirascope prompt file.

    The call to `ast.parse()` returns Python code as an AST, whereby each visitor method
    will be called for the corresponding nodes in the AST via `NodeVisitor.visit()`.

    Example:

    ```python
    analyzer = PromptAnalyzer()
    tree = ast.parse(file.read())
    analyzer.visit(tree)
    ```

    """

    def __init__(self) -> None:
        """Initializes the PromptAnalyzer."""
        self.imports: list[tuple[str, Optional[str]]] = []
        self.from_imports: list[tuple[str, str, Optional[str]]] = []
        self.variables: dict[str, Any] = {}
        self.classes: list[ClassInfo] = []
        self.functions: list[FunctionInfo] = []
        self.comments: str = ""

    def visit_Import(self, node) -> None:
        """Extracts imports from the given node."""
        for alias in node.names:
            self.imports.append((alias.name, alias.asname))
        self.generic_visit(node)

    def visit_ImportFrom(self, node) -> None:
        """Extracts from imports from the given node."""
        for alias in node.names:
            self.from_imports.append((node.module, alias.name, alias.asname))
        self.generic_visit(node)

    def visit_Assign(self, node) -> None:
        """Extracts variables from the given node."""
        target = node.targets[0]
        if isinstance(target, ast.Name):
            self.variables[target.id] = ast.unparse(node.value)
        self.generic_visit(node)

    def visit_ClassDef(self, node) -> None:
        """Extracts classes from the given node."""
        class_info = ClassInfo(
            name=node.name,
            bases=[ast.unparse(base) for base in node.bases],
            body="",
            decorators=[ast.unparse(decorator) for decorator in node.decorator_list],
            docstring=None,
        )

        # Extract docstring if present
        docstring = ast.get_docstring(node, False)
        if docstring:
            class_info.docstring = docstring

        # Handle the rest of the class body
        body_nodes = [n for n in node.body if not isinstance(n, ast.Expr)]
        body = []
        for node in body_nodes:
            if (
                isinstance(node, ast.Assign)
                and isinstance(node.targets[0], ast.Name)
                and node.targets[0].id == "prompt_template"
                and isinstance(node.value, ast.Constant)
                and node.end_lineno is not None
                and node.lineno < node.end_lineno
            ):
                # reconstruct template strings to be multi-line
                lines = node.value.s.split("\n")
                body.append(f'{node.targets[0].id} = """{lines.pop(0).strip()}')
                for i, line in enumerate(lines):
                    stripped_line = line.strip()
                    if stripped_line or i < len(lines) - 1:
                        body.append(line.strip())
                body.append('"""')
                body.append("")  # adds final newline
            else:
                body.append(ast.unparse(node))
        class_info.body = "\n".join(body)

        self.classes.append(class_info)

    def visit_AsyncFunctionDef(self, node):
        """Extracts async functions from the given node."""
        return self._visit_Function(node, is_async=True)

    def visit_FunctionDef(self, node):
        """Extracts functions from the given node."""
        return self._visit_Function(node, is_async=False)

    def _visit_Function(self, node, is_async):
        """Extracts functions or async functions from the given node."""
        # Initial function information setup
        function_info = FunctionInfo(
            name=node.name,
            args=[ast.unparse(arg) for arg in node.args.args],
            returns=ast.unparse(node.returns) if node.returns else None,
            body="",
            decorators=[ast.unparse(decorator) for decorator in node.decorator_list],
            docstring=None,
            is_async=is_async,  # Indicates whether the function is async
        )

        # Extract docstring if present
        docstring = ast.get_docstring(node, False)
        if docstring:
            function_info.docstring = docstring

        # Handle the rest of the function body
        body_nodes = [n for n in node.body if not isinstance(n, ast.Expr)]
        function_info.body = "\n".join(ast.unparse(n) for n in body_nodes)

        # Assuming you have a list to store functions
        self.functions.append(function_info)

    def visit_Module(self, node) -> None:
        """Extracts comments from the given node."""
        comments = ast.get_docstring(node, False)
        self.comments = "" if comments is None else comments
        self.generic_visit(node)

    def check_function_changed(self, other: PromptAnalyzer) -> bool:
        """Compares the functions of this file with those of another file."""
        return self._check_definition_changed(other, "function")

    def check_class_changed(self, other: PromptAnalyzer) -> bool:
        """Compares the classes of this file with those of another file."""
        return self._check_definition_changed(other)

    def _check_definition_changed(
        self,
        other: PromptAnalyzer,
        definition_type: Optional[Literal["class", "function"]] = "class",
    ) -> bool:
        """Compares classes or the functions of this file with those of another file"""

        self_definitions: Union[list[ClassInfo], list[FunctionInfo]] = (
            self.classes if definition_type == "class" else self.functions
        )
        other_definitions: Union[list[ClassInfo], list[FunctionInfo]] = (
            other.classes if definition_type == "class" else other.functions
        )

        self_definitions_dict = {
            definition.name: definition for definition in self_definitions
        }
        other_definitions_dict = {
            definition.name: definition for definition in other_definitions
        }

        all_definition_names = set(self_definitions_dict.keys()) | set(
            other_definitions_dict.keys()
        )

        for name in all_definition_names:
            if name in self_definitions_dict and name in other_definitions_dict:
                self_def_dict = self_definitions_dict[name].__dict__
                other_def_dict = other_definitions_dict[name].__dict__
                # Compare attributes of definitions with the same name
                def_diff = {
                    attr: (self_def_dict[attr], other_def_dict[attr])
                    for attr in self_def_dict
                    if self_def_dict[attr] != other_def_dict[attr]
                }
                if def_diff:
                    return True
            else:
                return True

        return False


def get_user_mirascope_settings(
    ini_file_path: str = "mirascope.ini",
) -> MirascopeSettings:
    """Returns the user's mirascope settings.

    Args:
        ini_file_path: The path to the mirascope.ini file.

    Returns:
        The user's mirascope settings as a `MirascopeSettings` instance.

    Raises:
        FileNotFoundError: If the mirascope.ini file is not found.
        KeyError: If the [mirascope] section is missing from the mirascope.ini file.
    """
    config = ConfigParser(allow_no_value=True)
    try:
        read_ok = config.read(ini_file_path)
        if not read_ok:
            raise FileNotFoundError(
                "The mirascope.ini file was not found. Please run "
                "`mirascope init` to create one or run the mirascope CLI from the "
                "same directory as the mirascope.ini file."
            )
        mirascope_config = config["mirascope"]
        return MirascopeSettings(**mirascope_config)
    except MissingSectionHeaderError as e:
        raise MissingSectionHeaderError(ini_file_path, e.lineno, e.source) from e


def prompts_directory_files() -> list[str]:
    """Returns a list of files in the user's prompts directory."""
    mirascope_settings = get_user_mirascope_settings()
    prompt_file_names = find_file_names(mirascope_settings.prompts_location)
    return [f"{name[:-3]}" for name in prompt_file_names]  # remove .py extension


def parse_prompt_file_name(prompt_file_name: str) -> str:
    """Returns the file name without the .py extension."""
    if prompt_file_name.endswith(".py"):
        return prompt_file_name[:-3]
    return prompt_file_name


def get_prompt_versions(version_file_path: str) -> VersionTextFile:
    """Returns the versions of the given prompt.

    Args:
        version_file_path: The path to the prompt.

    Returns:
        A `VersionTextFile` instance with the versions of current and latest revisions.
    """
    versions = VersionTextFile()
    try:
        with open(version_file_path, "r", encoding="utf-8") as file:
            file.seek(0)
            for line in file:
                # Check if the current line contains the key
                if line.startswith(CURRENT_REVISION_KEY + "="):
                    versions.current_revision = line.split("=")[1].strip()
                elif line.startswith(LATEST_REVISION_KEY + "="):
                    versions.latest_revision = line.split("=")[1].strip()
            return versions
    except FileNotFoundError:
        return versions


def check_prompt_changed(file1_path: Optional[str], file2_path: Optional[str]) -> bool:
    """Compare two prompts to check if the given prompts have changed.

    Args:
        file1_path: The path to the first prompt.
        file2_path: The path to the second prompt.

    Returns:
        Whether there are any differences between the two prompts.
    """
    if file1_path is None or file2_path is None:
        raise FileNotFoundError("Prompt or version file is missing.")
    # Parse the first file
    try:
        with open(file1_path, "r", encoding="utf-8") as file:
            content = file.read()
    except FileNotFoundError as e:
        raise FileNotFoundError(f"The file {file1_path} was not found.") from e
    analyzer1 = PromptAnalyzer()
    tree1 = ast.parse(content)
    analyzer1.visit(tree1)

    # Parse the second file
    try:
        with open(file2_path, "r", encoding="utf-8") as file:
            content = file.read()
    except FileNotFoundError as e:
        raise FileNotFoundError(f"The file {file2_path} was not found.") from e
    analyzer2 = PromptAnalyzer()
    tree2 = ast.parse(content)
    analyzer2.visit(tree2)
    # Compare the contents of the two files
    differences = {
        "comments": analyzer1.comments != analyzer2.comments,
        "imports_diff": bool(set(analyzer1.imports) ^ set(analyzer2.imports)),
        "from_imports_diff": bool(
            set(analyzer1.from_imports) ^ set(analyzer2.from_imports)
        ),
        "functions_diff": analyzer1.check_function_changed(analyzer2),
        "variables_diff": set(analyzer1.variables.keys()) - ignore_variables
        ^ set(analyzer2.variables.keys()) - ignore_variables,
        "classes_diff": analyzer1.check_class_changed(analyzer2),
        # Add other comparisons as needed
    }
    return any(differences.values())


def find_file_names(directory: str, prefix: str = "") -> list[str]:
    """Finds all files in a directory.

    Args:
        directory: The directory to search for the prompt.
        prefix: The prefix of the prompt to search for.

    Returns:
        A list of file names found.
    """
    pattern = os.path.join(directory, f"[!_]{prefix}*.py")  # ignores private files
    matching_files_with_dir = glob.glob(pattern)

    # Removing the directory part from each path
    return [os.path.basename(file) for file in matching_files_with_dir]


def find_prompt_paths(directory: Union[Path, str], prefix: str) -> Optional[list[str]]:
    """Finds and opens all prompts with the given directory.

    Args:
        directory: The directory to search for the prompt.
        prefix: The prefix of the prompt to search for.

    Returns:
        A list of paths to the prompt.
    """
    pattern = os.path.join(directory, prefix + "*.py")
    prompt_files = glob.glob(pattern)

    if not prompt_files:
        return None  # No files found

    # Return first file found
    return prompt_files


def find_prompt_path(directory: Union[Path, str], prefix: str) -> Optional[str]:
    """Finds and opens the first found prompt with the given directory.

    Args:
        directory: The directory to search for the prompt.
        prefix: The prefix of the prompt to search for.

    Returns:
        The path to the prompt.
    """
    prompt_files = find_prompt_paths(directory, prefix)
    if prompt_files:
        return prompt_files[0]
    return None


def get_prompt_analyzer(file: str) -> PromptAnalyzer:
    """Gets an instance of PromptAnalyzer for a file

    Args:
        file: The file to analyze

    Returns:
        An instance of PromptAnalyzer
    """
    analyzer = PromptAnalyzer()
    tree = ast.parse(file)
    analyzer.visit(tree)
    return analyzer


def _find_list_from_str(string: str) -> Optional[list[str]]:
    """Finds a list from a string.

    Args:
        string: The string to find the list from.

    Returns:
        The list found from the string.
    """
    start_bracket_index = string.find("[")
    end_bracket_index = string.find("]")
    if (
        start_bracket_index != -1
        and end_bracket_index != -1
        and end_bracket_index > start_bracket_index
    ):
        new_list = string[start_bracket_index : end_bracket_index + 1]
        return json.loads(new_list.replace("'", '"'))
    return None


def _update_tag_decorator_with_version(
    decorators: list[str], variables: MirascopeCliVariables, mirascope_alias: str
) -> Optional[str]:
    """Updates the tag decorator and returns the import name.

    Args:
        decorators: The decorators of the prompt.
        variables: The variables used by mirascope internal.
        mirascope_alias: The alias of the mirascope module.

    Returns:
        The import name of the tag decorator.
    """
    if variables.revision_id is None:
        return None
    import_name = "tags"
    tag_exists = False
    version_tag_prefix = "version:"  # mirascope tag prefix
    version_tag = f"{version_tag_prefix}{variables.revision_id}"
    for index, decorator in enumerate(decorators):
        if any(
            decorator.startswith(prefix)
            for prefix in ("tags(", f"{mirascope_alias}.tags(")
        ):
            tag_exists = True
            import_name = decorator.split("(")[0]
            decorator_arguments = _find_list_from_str(decorator)
            if decorator_arguments is not None:
                if f"{version_tag}" in decorator_arguments:
                    # The version tag already exists
                    break
                elif any(
                    argument.startswith(version_tag_prefix)
                    for argument in decorator_arguments
                ):
                    # Replace the version tag with the current version
                    for i, word in enumerate(decorator_arguments):
                        if word.startswith(version_tag_prefix):
                            decorator_arguments[i] = f"{version_tag}"
                    decorators[index] = f"{import_name}({decorator_arguments})"
                else:
                    # Tag decorator exists, append the current version to the tags
                    decorator_arguments.append(f"{version_tag}")
                    decorators[index] = f"{import_name}({decorator_arguments})"
            else:
                # Add tags decorator
                decorators[index] = f"{import_name}({version_tag})"
            break
    if not tag_exists:
        decorators.append(f'{import_name}(["{version_tag}"])')
    return import_name


def _update_mirascope_imports(imports: list[tuple[str, Optional[str]]]):
    """Updates the mirascope import.

    Args:
        imports: The imports from the PromptAnalyzer class
    """
    if not any(import_name == "mirascope" for import_name, _ in imports):
        imports.append(("mirascope", None))


def _update_mirascope_from_imports(
    member: str, from_imports: list[tuple[str, str, Optional[str]]]
):
    """Updates the mirascope from imports.

    Args:
        member: The member to import.
        from_imports: The from imports from the PromptAnalyzer class
    """
    if not any(
        (
            module_name == "mirascope"
            or module_name == "mirascope.base"
            or module_name == "mirascope.prompts"
        )
        and import_name == member
        for module_name, import_name, _ in from_imports
    ):
        from_imports.append(("mirascope", member, None))


def write_prompt_to_template(
    file: str,
    command: Literal[MirascopeCommand.ADD, MirascopeCommand.USE],
    variables: Optional[MirascopeCliVariables] = None,
) -> str:
    """Writes the given prompt to the template.

    Deconstructs a prompt with ast and reconstructs it using the Jinja2 template, adding
    revision history into the prompt when the command is `MirascopeCommand.ADD`.

    Args:
        file: The path to the prompt.
        command: The CLI command to execute.
        variables: A dictionary of revision ids which are rendered together with
            variable assignments that are not inside any class. Only relevant when
            `command` is `MirascopeCommand.ADD` - if `command` is
            `MirascopeCommand.USE`, `variables` should be `None`.

    Returns:
        The reconstructed prompt.
    """
    mirascope_settings = get_user_mirascope_settings()
    mirascope_directory = mirascope_settings.mirascope_location
    auto_tag = mirascope_settings.auto_tag
    template_loader = FileSystemLoader(searchpath=mirascope_directory)
    template_env = Environment(loader=template_loader)
    template = template_env.get_template("prompt_template.j2")
    analyzer = get_prompt_analyzer(file)
    if variables is None:
        variables = MirascopeCliVariables()

    if command == MirascopeCommand.ADD:
        # double quote revision ids to match how `ast.unparse()` formats strings
        new_variables = {
            k: f"'{v}'" if isinstance(v, str) else None
            for k, v in variables.__dict__.items()
        } | analyzer.variables
    else:  # command == MirascopeCommand.USE
        ignore_variable_keys = dict.fromkeys(ignore_variables, None)
        new_variables = {
            k: analyzer.variables[k]
            for k in analyzer.variables
            if k not in ignore_variable_keys
        }

    if auto_tag:
        import_tag_name: Optional[str] = None
        mirascope_alias = "mirascope"
        for name, alias in analyzer.imports:
            if name == "mirascope" and alias is not None:
                mirascope_alias = alias
                break
        for module, name, alias in analyzer.from_imports:
            if module == "mirascope" and name == "tags" and alias is not None:
                mirascope_alias = alias
                break

        for python_class in analyzer.classes:
            decorators = python_class.decorators
            if python_class.bases and python_class.bases[0] in mirascope_prompt_bases:
                import_tag_name = _update_tag_decorator_with_version(
                    decorators, variables, mirascope_alias
                )

        if import_tag_name == "tags":
            _update_mirascope_from_imports(import_tag_name, analyzer.from_imports)
        elif import_tag_name == f"{mirascope_alias}.tags":
            _update_mirascope_imports(analyzer.imports)

    data = {
        "comments": analyzer.comments,
        "variables": new_variables,
        "imports": analyzer.imports,
        "from_imports": analyzer.from_imports,
        "classes": analyzer.classes,
    }
    return template.render(**data)


def update_version_text_file(
    version_file: str,
    updates: dict[str, str],
) -> None:
    """Updates the version text file.

    Depending on the contents of `updates`, updates the ids of the current and latest
    revisions of the prompt.

    Args:
        version_file: The path to the version text file.
        updates: A dictionary containing updates to the current revision id and/or
            the latest revision id.
    """
    modified_lines = []
    edits_made = {
        key: False for key in updates
    }  # Track which keys already exist in the file
    version_file_path: Path = Path(version_file)
    if not version_file_path.is_file():
        version_file_path.touch()
    # Read the file and apply updates
    with open(version_file_path, "r", encoding="utf-8") as file:
        for line in file:
            # Check if the current line contains any of the keys
            for key, value in updates.items():
                if line.startswith(key + "="):
                    modified_lines.append(f"{key}={value}\n")
                    edits_made[key] = True
                    break
            else:
                # No key found, so keep the line as is
                modified_lines.append(line)

        # Add any keys that were not found at the end of the file
        for key, value in updates.items():
            if not edits_made[key]:
                modified_lines.append(f"{key}={value}\n")

    # Write the modified content back to the file
    with open(version_file_path, "w", encoding="utf-8") as file:
        file.writelines(modified_lines)


def check_status(
    mirascope_settings: MirascopeSettings, directory: str
) -> Optional[str]:
    """Checks the status of the given directory.

    Args:
        mirascope_settings: The user's mirascope settings.
        directory: The name of the prompt file (excluding the .py extension).

    Returns:
        The path to the prompt if the prompt has changed, otherwise `None`.
    """
    version_directory_path = mirascope_settings.versions_location
    prompt_directory_path = mirascope_settings.prompts_location
    version_file_name = mirascope_settings.version_file_name
    prompt_directory = os.path.join(version_directory_path, directory)
    used_prompt_path = f"{prompt_directory_path}/{directory}.py"

    # Get the currently used prompt version
    versions = get_prompt_versions(f"{prompt_directory}/{version_file_name}")
    current_head = versions.current_revision
    if current_head is None:
        return used_prompt_path
    current_version_prompt_path = find_prompt_path(prompt_directory, current_head)
    # Check if users prompt matches the current prompt version
    has_file_changed = check_prompt_changed(
        current_version_prompt_path, used_prompt_path
    )
    if has_file_changed:
        return used_prompt_path
    return None


def run_format_command(file: str) -> None:
    """Runs the format command on the given file.

    Args:
        file: The file to format
    """
    mirascope_settings = get_user_mirascope_settings()
    if mirascope_settings.format_command:
        format_commands: list[list[str]] = [
            command.split() for command in mirascope_settings.format_command.split(";")
        ]
        # assuming the final command takes filename as argument, and as final argument
        format_commands[-1].append(file)
        for command in format_commands:
            subprocess.run(
                [sys.executable, "-m"] + command,
                check=True,
                capture_output=True,
                shell=False,
            )



"""A module for interacting with OpenAI models."""
from contextlib import suppress

with suppress(ImportError):
    from . import wandb

from .calls import OpenAICall
from .extractors import OpenAIExtractor
from .tools import OpenAITool
from .types import OpenAICallParams, OpenAICallResponse, OpenAICallResponseChunk



"""The init command for the Mirascope CLI."""
import os
from importlib.resources import files
from pathlib import Path

from jinja2 import Template
from typer import Option

from ..schemas import MirascopeSettings


def init_command(
    mirascope_location: str = Option(
        help="Main mirascope directory", default=".mirascope"
    ),
    prompts_location: str = Option(
        help="Location of prompts directory", default="prompts"
    ),
) -> None:
    """Initializes the mirascope project.

    Creates the project structure and files needed for mirascope to work.

    Initial project structure:
    ```
    |
    |-- mirascope.ini
    |-- .mirascope
    |   |-- prompt_template.j2
    |   |-- versions/
    |   |   |-- <directory_name>/
    |   |   |   |-- version.txt
    |   |   |   |-- <revision_id>_<directory_name>.py
    |-- prompts/
    ```

    Args:
        mirascope_location: The root mirascope directory to create.
        prompts_location: The user's prompts directory.
    """
    destination_dir = Path.cwd()
    versions_directory = os.path.join(mirascope_location, "versions")
    os.makedirs(versions_directory, exist_ok=True)
    print(f"Creating {destination_dir}/{versions_directory}")
    os.makedirs(prompts_location, exist_ok=True)
    print(f"Creating {destination_dir}/{prompts_location}")
    prompts_init_file: Path = Path(f"{destination_dir}/{prompts_location}/__init__.py")
    if not prompts_init_file.is_file():
        prompts_init_file.touch()
        print(f"Creating {prompts_init_file}")
    # Create the 'mirascope.ini' file in the current directory with some default values
    ini_settings = MirascopeSettings(
        mirascope_location=mirascope_location,
        versions_location="versions",
        prompts_location=prompts_location,
        version_file_name="version.txt",
        auto_tag=True,
    )

    # Get templates from the mirascope.cli.generic package
    generic_file_path = files("mirascope.cli.generic")
    ini_path = generic_file_path.joinpath("mirascope.ini.j2")
    with open(str(ini_path), "r", encoding="utf-8") as file:
        template = Template(file.read())
        rendered_content = template.render(ini_settings.model_dump())
        destination_file_path = destination_dir / "mirascope.ini"
        with open(destination_file_path, "w", encoding="utf-8") as destination_file:
            destination_file.write(rendered_content)
            print(f"Creating {destination_file_path}")

    # Create the 'prompt_template.j2' file in the mirascope directory specified by user
    prompt_template_path = generic_file_path.joinpath("prompt_template.j2")
    with open(str(prompt_template_path), "r", encoding="utf-8") as file:
        content = file.read()
    template_path = os.path.join(mirascope_location, "prompt_template.j2")
    with open(template_path, "w", encoding="utf-8") as file:
        file.write(content)
        print(f"Creating {destination_dir}/{template_path}")

    print("Initialization complete.")
